from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client


class Segment(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.field_name = data["field_name"]
        self.values = data["values"]
        self.term = data["term"]

    @classmethod
    def get_all(
        cls, client: Client, model_id: Optional[str] = None, manual_only: bool = True
    ) -> List["Segment"]:
        url_params = {}
        if manual_only:
            url_params["is_manually_created"] = "true"
        if model_id is not None:
            url_params["model_id"] = model_id
        encoded_url_params = urlencode(url_params)
        response = client.send_request(
            f"/data-segments{f'?{encoded_url_params}' if len(url_params) > 0 else f''}",
            "GET",
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        model_id: str,
        field_name: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        terms: Optional[List[Tuple[str, str]]] = None,
    ) -> "Segment":
        segment_data = {}
        if terms is not None:
            segment_data["terms_values"] = terms
        else:
            segment_data["field_name"] = field_name
            segment_data["values"] = values
        response = client.send_request(
            "/data-segments",
            "POST",
            {"name": name, "model_id": model_id, **segment_data},
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Segment":
        response = client.send_request(f"/data-segments/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(
        self,
        name: Optional[str] = None,
        # field_name: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        term: Optional[str] = None,
        **kwargs,
    ):
        args = {}
        if name is not None:
            args["name"] = name
        # if field_name is not None:
        #     args["field_name"] = field_name
        if values is not None:
            args["values"] = values
        if term is not None:
            args["term"] = term
        response = self.client.send_request(
            f"/data-segments/{self.id}",
            "POST",
            args,
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        # response = self.client.send_request(f"/data-segments/{self.id}", "PUT", args)
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/data-segments/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/data-segments/{id}", "DELETE")
        client.assert_response(response)



from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
import json
import pathlib
import string
import traceback
from typing import Any, Dict, List, Optional, Tuple

from aporia.as_code.resources.base import BaseResource, CompareStatus, NonDeletableResourceException
from aporia.as_code.resources.custom_metrics import CustomMetric
from aporia.as_code.resources.data_source import DataSource
from aporia.as_code.resources.dataset import Dataset
from aporia.as_code.resources.model import Model
from aporia.as_code.resources.monitor import Monitor
from aporia.as_code.resources.segment import Segment
from aporia.as_code.resources.version import Version
from aporia.sdk.aporia_api import Aporia
from aporia.sdk.aporia_api import BaseAporiaResource as SDKBaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.custom_metrics import CustomMetric as SDKCustomMetric
from aporia.sdk.datasets import Dataset as SDKDataset
from aporia.sdk.models import Model as SDKModel
from aporia.sdk.models import ModelColor
from aporia.sdk.monitors import (
    ActionType,
    infer_detection_method,
    LogicEvaluationType,
    MessagingIntegrationType,
    MetricType,
)
from aporia.sdk.monitors import Monitor as SDKMonitor
from aporia.sdk.monitors import MonitorType, parse_monitor_configuration, PreConditionType
from aporia.sdk.segments import Segment as SDKSegment
from aporia.sdk.versions import Version as SDKVersion


class StackAction(Enum):
    CREATE = "create"
    DELETE = "delete"
    PREVIEW = "preview"
    REFRESH = "refresh"
    DIFF = "diff"


class ResourceStatus(Enum):
    CREATED = "created"  # Found in config file
    MISSING = "missing"  # Not found in config file, create
    DEPRECATED = "deprecated"  # Found in config file, definition changed
    # TODO: Support that
    # UNWANTED = "unwanted"  # Found in config file, not in resources


@dataclass
class Resource:
    resource: BaseResource
    depth: int
    status: ResourceStatus
    id: Optional[str] = None


RESOURCE_TYPE_TO_RESOURCE_CLASS = {
    CustomMetric.__name__: CustomMetric,
    DataSource.__name__: DataSource,
    Dataset.__name__: Dataset,
    Model.__name__: Model,
    Monitor.__name__: Monitor,
    Segment.__name__: Segment,
    Version.__name__: Version,
}


class Stack:
    def __init__(
        self,
        token: str,
        account: str,
        workspace: str = "default-workspace",
        base_url: str = "https://platform.aporia.com",
    ):
        self._resources: OrderedDict[str, Resource] = OrderedDict()
        self._output = None
        self._base_url = base_url
        self._client = Client(base_url=f"{base_url}/api/v1/{account}/{workspace}", token=token)
        self._delete_resources: OrderedDict[str, BaseResource] = OrderedDict()
        self._aporia_api = Aporia(
            token=token, account_name=account, workspace_name=workspace, base_url=base_url
        )

    def add(self, resource: BaseResource, depth=0, prefix=None):
        unique_id = (prefix or "") + resource.name

        if unique_id in self._resources.keys():
            raise Exception(f"Found duplicate resource name: {unique_id}")

        self._resources[unique_id] = Resource(
            resource=resource, depth=depth, status=ResourceStatus.MISSING
        )
        if hasattr(resource, "sub_resources"):
            for sub_resource, _ in resource.sub_resources:
                self.add(sub_resource, depth=depth + 1, prefix=f"{prefix or ''}{resource.name}/")

    def _call_argsetters(self, resource, resource_data):
        if hasattr(resource, "sub_resources"):
            sub_resources = resource.sub_resources
            for sub_resource, argsetter in sub_resources:
                argsetter(resource_data, sub_resource)
        if hasattr(resource, "dependants"):
            dependants = resource.dependants
            for sub_resource, argsetter in dependants:
                argsetter(resource_data, sub_resource)

    def _preview(self, delete: bool = False):
        if delete:
            RESOURCE_STATUS_TO_PREVIEW_ACTION = {
                ResourceStatus.CREATED: "delete",
                ResourceStatus.MISSING: "create",
                ResourceStatus.DEPRECATED: "delete",
            }
            for resource_entry in self._resources.values():
                if resource_entry.status is ResourceStatus.MISSING:
                    continue
                print(
                    f"|-{'----' * resource_entry.depth}{resource_entry.resource.name} - {RESOURCE_STATUS_TO_PREVIEW_ACTION[resource_entry.status]}"
                )
        else:
            RESOURCE_STATUS_TO_PREVIEW_ACTION = {
                ResourceStatus.CREATED: "pass",
                ResourceStatus.MISSING: "create",
                ResourceStatus.DEPRECATED: "update",
            }
            for resource_entry in self._resources.values():
                print(
                    f"|-{'----' * resource_entry.depth}{resource_entry.resource.name} - {RESOURCE_STATUS_TO_PREVIEW_ACTION[resource_entry.status]}"
                )
        if len(self._delete_resources) > 0:
            print("Removed resources:")
        for unique_id in self._delete_resources.keys():
            resource_parts = unique_id.split("/")
            depth = len(resource_parts) - 1
            name = resource_parts[-1]
            print(f"|-{'----' * depth}{name} - delete")

    def _get_diff_config(self, output_format: str = "print"):
        diffs = {}
        for unique_id, resource_entry in self._resources.items():
            if resource_entry.status is not ResourceStatus.DEPRECATED:
                continue
            resource_data = self._output[unique_id]
            resource_diff = resource_entry.resource.get_diff(resource_data=resource_data)
            diffs[unique_id] = {}
            for config_name, (old, new) in resource_diff.items():
                diffs[unique_id][config_name] = {"old": old, "new": new}
        if output_format == "json":
            print(json.dumps(diffs))
        elif output_format == "print":
            for unique_id, diffed_configs in diffs.items():
                print(f"Resource '{unique_id}' changed. Changes:")
                for config_name, config_diffs in diffed_configs.items():
                    print(
                        f"    Configuration '{config_name}' changed.\n        Old: {config_diffs['old']}\n        New: {config_diffs['new']}"
                    )

    def __pop_deleted_sub_resources(self, non_deleteable_resources: List[str], unique_id):
        resource_entry = self._resources[unique_id]
        if hasattr(resource_entry.resource, "sub_resources"):
            for resource, _ in resource_entry.resource.sub_resources:
                sub_resource_unique_id = f"{unique_id}/{resource.name}"
                if sub_resource_unique_id in non_deleteable_resources:
                    self.__pop_deleted_sub_resources(
                        non_deleteable_resources=non_deleteable_resources,
                        unique_id=sub_resource_unique_id,
                    )
                    self._output.pop(sub_resource_unique_id)
                elif sub_resource_unique_id in self._output.keys():
                    raise RuntimeError(
                        f"Resource deletion missed for {sub_resource_unique_id}. Stack potentially corrupted."
                    )

    def _create(self) -> Tuple[Dict, Optional[Exception]]:
        output = self._output or {}
        error = None
        try:
            # Create/update resources
            for (
                unique_id,
                resource_entry,
            ) in self._resources.items():
                if resource_entry.status is ResourceStatus.CREATED:
                    continue
                if hasattr(resource_entry.resource, "deferred_load"):
                    resource_entry.resource.deferred_load(self._client, resource_entry.resource)

                if resource_entry.status is ResourceStatus.MISSING:
                    _, resource_data = resource_entry.resource.create(client=self._client)
                elif resource_entry.status is ResourceStatus.DEPRECATED:
                    resource_data = resource_entry.resource.update(
                        client=self._client, id=self._output[unique_id]["id"]
                    )
                else:
                    raise Exception("Unexpected resource status")
                self._call_argsetters(resource_entry.resource, resource_data)
                output[unique_id] = resource_data
                output[unique_id]["_resource_type"] = type(resource_entry.resource).__name__
                resource_entry.status = ResourceStatus.CREATED

            # Delete resources
            non_deleteable_resources = []
            for unique_id, resource_class in list(self._delete_resources.items())[::-1]:
                resource_data = self._output[unique_id]
                try:
                    resource_class.delete(client=self._client, id=resource_data["id"])
                except NonDeletableResourceException:
                    non_deleteable_resources.append(unique_id)
                    continue
                for non_deleteable_resource in [*non_deleteable_resources]:
                    if non_deleteable_resource.startswith(f"{unique_id}/"):
                        self._output.pop(non_deleteable_resource)
                        non_deleteable_resources.remove(non_deleteable_resource)
                self._output.pop(unique_id)
        except Exception as e:
            error = e

        return output, error

    def _destroy(self):
        # Read from storage
        # Destroy all resources
        # TODO: Consider adding all this logic into _create, just with setting resource.action to delete
        error = None
        if self._output is None:
            raise Exception("Stack not created")
        try:
            non_deleteable_resources = []
            for unique_id, resource_entry in list(self._resources.items())[::-1]:
                if unique_id in self._output.keys():
                    resource_data = self._output[unique_id]
                    try:
                        resource_entry.resource.delete(client=self._client, id=resource_data["id"])
                    except NonDeletableResourceException:
                        non_deleteable_resources.append(unique_id)
                        continue
                    self.__pop_deleted_sub_resources(non_deleteable_resources, unique_id)
                    self._output.pop(unique_id)
            for unique_id, resource_class in list(self._delete_resources.items())[::-1]:
                resource_data = self._output[unique_id]
                try:
                    resource_class.delete(client=self._client, id=resource_data["id"])
                except NonDeletableResourceException:
                    non_deleteable_resources.append(unique_id)
                    continue
                self.__pop_deleted_sub_resources(non_deleteable_resources, unique_id)
                self._output.pop(unique_id)
        except Exception as e:
            error = e

        return error

    def _diff(self):
        for unique_id, resource_entry in self._resources.items():
            if unique_id in self._output.keys():
                # Resource already exists. Call argsetters
                # TODO: Add support for edit
                resource_data = self._output[unique_id]
                self._call_argsetters(resource_entry.resource, resource_data)
                compare_status = self._resources[unique_id].resource.compare(
                    resource_data=resource_data
                )
                if compare_status is CompareStatus.SAME:
                    self._resources[unique_id].status = ResourceStatus.CREATED
                elif compare_status is CompareStatus.UPDATEABLE:
                    self._resources[unique_id].status = ResourceStatus.DEPRECATED
                else:
                    raise Exception(
                        f"Configuration for {unique_id} was edited but can't be updated!"
                    )
                continue

        for unique_id, resource_data in self._output.items():
            if unique_id not in self._resources.keys():
                self._delete_resources[unique_id] = RESOURCE_TYPE_TO_RESOURCE_CLASS[
                    resource_data["_resource_type"]
                ]

    def __resource_name_to_python_name(self, name: str) -> str:
        result = "".join(
            [c if c in (string.ascii_letters + string.digits + "_") else "_" for c in name]
        ).lower()
        if result[0] in string.digits:
            result = f"_{result}"
        return result

    def __create_python_representation(
        self,
        resource: SDKBaseAporiaResource,
        index: int,
        version_id_to_object: Optional[Dict[str, str]] = None,
        segment_id_to_object: Optional[Dict[str, str]] = None,
        custom_metric_id_to_object: Optional[Dict[str, str]] = None,
    ) -> Tuple[str, str]:
        if isinstance(resource, SDKSegment):
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Segment(
    "{resource.name}",
"""
            if resource.raw_data["field"] is not None:
                code += f"""\tfield="{resource.raw_data["field"]["name"]}",\n"""
                code += f"""\tvalues={resource.raw_data["values"]}\n"""
            else:
                code += f"""\tterms={resource.raw_data["terms_values"]}\t"""
            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKCustomMetric):
            return f'''{self.__resource_name_to_python_name(resource.name)} = aporia.CustomMetric(
    "{resource.name}",
    code="""{resource.raw_data["code"]}""",
)''', self.__resource_name_to_python_name(
                resource.name
            )
        elif isinstance(resource, SDKDataset):
            schema = resource.raw_data["schema"]
            raw_inputs = {
                raw_input["name"]: raw_input["type"] for raw_input in schema["raw_inputs"]
            }
            features = {feature["name"]: feature["type"] for feature in schema["features"]}
            predictions = {
                prediction["name"]: prediction["type"] for prediction in schema["predictions"]
            }
            actuals = {actual["name"]: actual["type"] for actual in schema["actuals"]}
            actual_mappings = {
                actual["name"]: actual["properties"]["prediction"] for actual in schema["actuals"]
            }
            code = f"""{resource.raw_data["stage"]}_dataset_{index} = aporia.Dataset(
    "{resource.raw_data["stage"]}",
    type="{resource.raw_data["stage"]}",
    data_source_id="{resource.raw_data["data_source_id"]}",
    connection_data={resource.raw_data["config"]},
"""
            if schema["id_column"] is not None:
                code += f"""\tid_column="{schema['id_column']}",\n"""
            if schema["timestamp_column"] is not None:
                code += f"""\ttimestamp_column="{schema['timestamp_column']}",\n"""
            if len(raw_inputs) > 0:
                code += f"\traw_inputs={raw_inputs},\n"
            if len(features) > 0:
                code += f"\tfeatures={features},\n"
            if len(predictions) > 0:
                code += f"\tpredictions={predictions},\n"
            if len(actuals) > 0:
                code += f"\tactuals={actuals},\n"
            if len(actual_mappings) > 0:
                code += f"\tactual_mappings={actual_mappings},\n"
            code += ")"
            return code, f'{resource.raw_data["stage"]}_dataset_{index}'
        elif isinstance(resource, SDKVersion):
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Version(
    "{resource.name}",
"""
            if resource.raw_data["serving_dataset"] is not None:
                code += f"\tserving=serving_dataset_{index},\n"
            if resource.raw_data["training_dataset"] is not None:
                code += f"\ttraining=training_dataset_{index},\n"
            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKMonitor):
            parsed_config = parse_monitor_configuration(resource.raw_data["configuration"])
            detection_method = infer_detection_method(parsed_config)
            code = f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Monitor(\n"""
            code += f'\t"{resource.name}",\n'
            code += f"\tmonitor_type=aporia.MonitorType.{resource.type.name},\n"
            code += f"\tdetection_method=aporia.DetectionMethod.{detection_method.name},\n"

            if parsed_config.metric.type == MetricType.EUCLIDEAN_DISTANCE:
                code += "\tis_embedding_monitor=True,\n"

            # Identification configuration
            if parsed_config.identification.raw_inputs is not None:
                raw_inputs = parsed_config.identification.raw_inputs
                raw_inputs_array = (
                    f"""[{', '.join(f'"{raw_input}"' for raw_input in raw_inputs)}]"""
                )
                code += f"\traw_inputs={raw_inputs_array},\n"
            if parsed_config.identification.features is not None:
                features = parsed_config.identification.features
                features_array = f"""[{', '.join(f'"{feature}"' for feature in features)}]"""
                code += f"\tfeatures={features_array},\n"
            if parsed_config.identification.predictions is not None:
                predictions = parsed_config.identification.predictions
                predictions_array = (
                    f"""[{', '.join(f'"{prediction}"' for prediction in predictions)}]"""
                )
                code += f"\tpredictions={predictions_array},\n"
            if parsed_config.identification.models.version is not None:
                version_id = parsed_config.identification.models.version
                if version_id in (version_id_to_object or {}):
                    code += f"\tversion={version_id_to_object[version_id]},\n"
                else:
                    code += f'\tversion_id="{version_id}",\n'

            if parsed_config.identification.segment.group is not None:
                segment_id = parsed_config.identification.segment.group
                if segment_id in (segment_id_to_object or {}):
                    code += f"\tsegment={segment_id_to_object[segment_id]},\n"
                else:
                    code += f'\tsegment_id="{segment_id}",\n'

            if parsed_config.identification.segment.value is not None:
                segment_value = parsed_config.identification.segment.value
                if isinstance(segment_value, str):
                    segment_value = f'"{segment_value}"'
                code += f"\tsegment_value={segment_value},\n"

            # Dataset configuration:
            code += f"\tfocal=aporia.FocalConfiguration(**{parsed_config.focal.to_dict()}),\n"
            if parsed_config.baseline is not None:
                code += f"\tbaseline=aporia.BaselineConfiguration(**{parsed_config.baseline.to_dict()}),\n"

            # Metric Configuration
            if resource.type in [MonitorType.PERFORMANCE_DEGRADATION, MonitorType.METRIC_CHANGE]:
                code += f"\tmetric=aporia.MetricType.{parsed_config.metric.type.name},\n"
            metric_config = parsed_config.metric
            if metric_config.id is not None:
                custom_metric_id = metric_config.id
                if custom_metric_id in (custom_metric_id_to_object or {}):
                    code += f"\tcustom_metric={custom_metric_id_to_object[custom_metric_id]},\n"
                else:
                    code += f'\tcustom_metric_id="{custom_metric_id}",\n'
            if metric_config.average is not None:
                code += f"\taverage_method=aporia.AverageMethod.{metric_config.average.name},\n"
            if metric_config.metricAtK is not None:
                code += f"\tk={metric_config.metricAtK},\n"
            if metric_config.threshold is not None:
                code += f"\tprediction_threshold={metric_config.threshold},\n"
            if metric_config.metricPerClass is not None:
                prediction_class = metric_config.metricPerClass
                if isinstance(prediction_class, str):
                    prediction_class = f'"{prediction_class}"'
                code += f"\tprediction_class={prediction_class},\n"

            # Logic Evaluation Configuration
            # TODO: Support multiple logic evaluations, if needed
            logic_evaluation_configuration = parsed_config.logicEvaluations[0]
            added_percentage = False
            use_percentage = False
            use_staleness = False
            if logic_evaluation_configuration.name is LogicEvaluationType.RATIO:
                use_percentage = True
            if logic_evaluation_configuration.name is LogicEvaluationType.MODEL_STALENESS:
                use_staleness = True
            if logic_evaluation_configuration.min is not None:
                if use_percentage:
                    if not added_percentage:
                        code += f"\tpercentage={(1-logic_evaluation_configuration.min)*100},\n"
                        added_percentage = True
                else:
                    code += f"\tmin={logic_evaluation_configuration.min},\n"
            if logic_evaluation_configuration.max is not None:
                if use_percentage:
                    if not added_percentage:
                        code += f"\tpercentage={(logic_evaluation_configuration.max-1)*100},\n"
                        added_percentage = True
                elif use_staleness:
                    code += (
                        f'\tstaleness_period="{logic_evaluation_configuration.max.to_string()}",\n'
                    )
                else:
                    code += f"\tmax={logic_evaluation_configuration.max},\n"
            if logic_evaluation_configuration.sensitivity is not None:
                code += f"\tsensitivity={logic_evaluation_configuration.sensitivity},\n"
            if logic_evaluation_configuration.testOnlyIncrease is not None:
                code += (
                    f"\talert_on_increase_only={logic_evaluation_configuration.testOnlyIncrease},\n"
                )
            if logic_evaluation_configuration.thresholds is not None:
                code += f"\tthresholds=aporia.ThresholdConfiguration(**{logic_evaluation_configuration.thresholds.to_dict()}),\n"
            if logic_evaluation_configuration.distance is not None:
                code += f"\tdistance={logic_evaluation_configuration.distance},\n"
            if logic_evaluation_configuration.new_values_count_threshold is not None:
                code += f"\tnew_values_count_threshold={logic_evaluation_configuration.new_values_count_threshold},\n"
            if logic_evaluation_configuration.new_values_ratio_threshold is not None:
                code += f"\tnew_values_ratio_threshold={logic_evaluation_configuration.new_values_ratio_threshold},\n"

            # Alert configuration
            added_alert = False
            for action in parsed_config.actions:
                if action.type is ActionType.ALERT:
                    if not added_alert:
                        emails = []
                        messaging = {}
                        code += f"\tseverity=aporia.Severity.{action.severity.name},\n"
                        for notification in action.notification:
                            if notification["type"] == "EMAIL":
                                emails.extend(notification["emails"])
                            else:
                                integration_type = (
                                    f"{MessagingIntegrationType(notification['type']).value}"
                                )
                                if integration_type not in messaging.keys():
                                    messaging[integration_type] = []
                                messaging[integration_type].append(notification["integration_id"])
                        if len(emails) > 0:
                            code += f"\temails={emails},\n"
                        if len(messaging) > 0:
                            code += f"\tmessaging={messaging},\n"
                        added_alert = True
                        if action.maxAlertsPerDay is not None:
                            code += f"\tmax_alerts_per_day={action.maxAlertsPerDay},\n"

            # Precondition Configuration
            added_min = False
            added_min_focal_datapoints = False
            added_min_baseline_datapoints = False
            for precondition in parsed_config.preConditions or []:
                if precondition.name is PreConditionType.FOCAL_DATA_VALUE_IN_RANGE:
                    if not added_min:
                        code += f"\tmin={precondition.min},\n"
                        added_min = True
                if precondition.name is PreConditionType.MIN_FOCAL_DATA_POINTS:
                    if not added_min_focal_datapoints:
                        code += f"\tmin_focal_prediction_count={precondition.value},\n"
                        added_min_focal_datapoints = True
                if precondition.name is PreConditionType.MIN_BASELINE_DATA_POINTS:
                    if not added_min_baseline_datapoints:
                        code += f"\tmin_baseline_prediction_count={precondition.value},\n"
                        added_min_baseline_datapoints = True

            code += ")"
            return code, self.__resource_name_to_python_name(resource.name)
        elif isinstance(resource, SDKModel):
            return f"""{self.__resource_name_to_python_name(resource.name)} = aporia.Model(
    "{resource.name}",
    type=aporia.ModelType.{resource.type.name},
    icon=aporia.ModelIcon.{resource.icon.name},
    color={f'aporia.ModelColor.{resource.color.name}' if isinstance(resource.color, ModelColor) else f'"{resource.color}"'},
    versions=[],
    segments=[],
    custom_metrics=[],
    monitors=[],
)""", self.__resource_name_to_python_name(
                resource.name
            )
        raise TypeError(f"Unsupported type: {type(resource).__name__})")

    def __create_model_code(
        self,
        model: SDKModel,
        versions: List[Tuple[SDKVersion, List[SDKDataset]]],
        segments: List[SDKSegment],
        custom_metrics: List[SDKCustomMetric],
        monitors: List[SDKMonitor],
    ) -> Tuple[str, str]:
        model_code, model_name = self.__create_python_representation(model, 0)
        model_versions = []
        model_segments = []
        model_custom_metrics = []
        model_monitors = []
        versions_code = []
        segments_code = []
        custom_metrics_code = []
        monitors_code = []
        version_id_to_object = {}
        segment_id_to_object = {}
        custom_metric_id_to_object = {}
        for index, (version, datasets) in enumerate(versions):
            datasets_code = [
                self.__create_python_representation(dataset, index=index)[0] for dataset in datasets
            ]
            version_code, version_name = self.__create_python_representation(version, index=index)
            version_id_to_object[version.id] = version_name
            full_version_code = "\n".join([*datasets_code, version_code])
            versions_code.append(full_version_code)
            model_versions.append(version_name)
        for index, segment in enumerate(segments):
            segment_code, segment_name = self.__create_python_representation(segment, index=index)
            segment_id_to_object[segment.id] = segment_name
            segments_code.append(segment_code)
            model_segments.append(segment_name)
        for index, custom_metric in enumerate(custom_metrics):
            custom_metric_code, custom_metric_name = self.__create_python_representation(
                custom_metric, index=index
            )
            custom_metric_id_to_object[custom_metric.id] = custom_metric_name
            custom_metrics_code.append(custom_metric_code)
            model_custom_metrics.append(custom_metric_name)
        for index, monitor in enumerate(monitors):
            monitor_code, monitor_name = self.__create_python_representation(
                monitor,
                index=index,
                version_id_to_object=version_id_to_object,
                segment_id_to_object=segment_id_to_object,
                custom_metric_id_to_object=custom_metric_id_to_object,
            )
            monitors_code.append(monitor_code)
            model_monitors.append(monitor_name)

        model_code = model_code.replace("versions=[]", f"versions=[{', '.join(model_versions)}]")
        model_code = model_code.replace("segments=[]", f"segments=[{', '.join(model_segments)}]")
        model_code = model_code.replace(
            "custom_metrics=[]", f"custom_metrics=[{', '.join(model_custom_metrics)}]"
        )
        model_code = model_code.replace("monitors=[]", f"monitors=[{', '.join(model_monitors)}]")

        # Hack to allow newlines in f-string expressions
        newline = "\n"
        full_code = """import aporia.as_code as aporia
"""
        full_code += f"""
# Creating Model {model.name}
"""
        if len(versions) > 0:
            full_code += f"""
## Define Versions
{newline.join(versions_code)}
"""
        if len(segments) > 0:
            full_code += f"""
## Define Segments
{newline.join(segments_code)}
"""

        if len(custom_metrics) > 0:
            full_code += f"""
## Define Custom Metrics
{newline.join(custom_metrics_code)}
"""

        if len(monitors) > 0:
            full_code += f"""
## Define Monitors
{newline.join(monitors_code)}
"""

        full_code += "## Creating model\n"
        full_code += model_code

        full_code += f"""

def add_model(stack: aporia.Stack):
    stack.add({model_name})
"""

        return full_code, model_name

    def __create_resource_entry(
        self, resource: SDKBaseAporiaResource, parent_unique_id: Optional[str] = None
    ) -> Tuple[str, Dict]:
        data = deepcopy(resource.raw_data)
        data["_resource_type"] = type(resource).__name__
        if isinstance(resource, SDKDataset):
            resource_name = resource.raw_data["stage"]
        else:
            resource_name = resource.name
        name = f"{parent_unique_id + '/' if parent_unique_id is not None else ''}{resource_name.replace('/', '-')}"
        return (name, data)

    def __iterate_model(
        self, model: SDKModel, index: int
    ) -> Tuple[Dict[str, Dict[str, Any]], str, str]:
        resources = OrderedDict()
        model_name, data = self.__create_resource_entry(resource=model)
        resources[model_name] = data

        model_versions = []
        for version in model.get_versions():
            datasets = version.get_datasets()
            model_versions.append((version, datasets))
        model_segments = model.get_segments()
        model_custom_metrics = model.get_custom_metrics()
        model_monitors = model.get_monitors()

        # Add versions
        for version, datasets in model_versions:
            version_name, data = self.__create_resource_entry(
                resource=version, parent_unique_id=model_name
            )
            resources[version_name] = data
            # Add datasets
            for dataset in datasets:
                dataset_name, data = self.__create_resource_entry(
                    resource=dataset, parent_unique_id=version_name
                )
                resources[dataset_name] = data

        # Add segments
        for segment in model_segments:
            segment_name, data = self.__create_resource_entry(
                resource=segment, parent_unique_id=model_name
            )
            resources[segment_name] = data

        # Add custom metrics
        for custom_metric in model_custom_metrics:
            custom_metric_name, data = self.__create_resource_entry(
                resource=custom_metric, parent_unique_id=model_name
            )
            resources[custom_metric_name] = data

        # Add custom metrics
        for monitor in model_monitors:
            monitor_name, data = self.__create_resource_entry(
                resource=monitor, parent_unique_id=model_name
            )
            resources[monitor_name] = data

        model_code, model_name = self.__create_model_code(
            model=model,
            versions=model_versions,
            segments=model_segments,
            custom_metrics=model_custom_metrics,
            monitors=model_monitors,
        )

        return resources, model_code, model_name

    def _import(self, output_directory: str):
        # data_sources = self._aporia_api.get_data_sources()
        models = self._aporia_api.get_models()
        resources = OrderedDict()
        output_path = pathlib.Path(output_directory)
        output_path.mkdir(parents=True, exist_ok=True)

        models_code = {}

        for index, model in enumerate(models):
            model_resources, model_code, model_name = self.__iterate_model(model=model, index=index)
            for name, data in model_resources.items():
                if name in resources.keys():
                    raise NameError(f"Resource {name} found duplicate!")
                resources[name] = data
            models_code[model_name] = model_code

        code = [
            "import os",
            "import aporia.as_code as aporia\n",
            "",
            'account = os.environ["APORIA_ACCOUNT"]',
            'workspace = os.environ["APORIA_WORKSPACE"]',
            'token = os.environ["APORIA_TOKEN"]',
            "",
            "stack = aporia.Stack(",
            "    account=account,",
            "    workspace=workspace,",
            "    token=token,",
            f'    base_url="{self._base_url}"',
            ")",
            "",
        ]

        code.append("\n# Registering Models")
        for model_name, model_code in models_code.items():
            with open(output_path / f"{model_name}.py", "w") as model_file:
                model_file.write(model_code)
                code.append(f"import {model_name}")
                code.append(f"{model_name}.add_model(stack)")

        code.append("\n# Applying Stack")
        code.append('stack.apply(config_path="stack.json", action="create", rollback=False)\n')

        with open(output_path / "stack.json", "w") as json_file:
            json_file.write(json.dumps(resources))

        with open(output_path / "main.py", "w") as python_file:
            python_file.write("\n".join(code))

    def apply(
        self,
        action: StackAction = StackAction.CREATE,
        skip_preview: bool = False,
        yes: bool = False,
        rollback=True,
        yes_rollback=False,
        config_path: Optional[str] = None,
        debug: Optional[bool] = False,
        diff_format: str = "print",
    ):
        action = StackAction(action)
        # TODO: Read config_path, and apply diffs. Then write to config_path
        if config_path is not None:
            try:
                with open(config_path, "r") as f:
                    self._output = json.loads(f.read())
                    self._diff()
            except FileNotFoundError:
                # It's OK if the file wasn't found. It's probably the first run
                pass

        if action is StackAction.CREATE:
            if not skip_preview:
                self._preview()
            if not yes:
                if input("Do create? [Y/N] ").lower().strip() not in ["y"]:
                    print("Cancelling...")
                    return
            self._output, error = self._create()

            if error is not None:
                if debug:
                    traceback.print_exception(error, tb=error.__traceback__, value=error)
                else:
                    print(error)
                if rollback:
                    if not yes_rollback:
                        if input("Delete created resources? [Y/N] ").lower().strip() in ["y"]:
                            print("Deleting...")
                            error = self._destroy()
                            if error is not None:
                                if debug:
                                    traceback.print_exception(error, tb=error.__traceback__, value=error)
                                else:
                                    print(error)
        elif action is StackAction.PREVIEW:
            self._preview()
            return
        elif action is StackAction.DELETE:
            if not skip_preview:
                self._preview(delete=True)
            # TODO: Change the way you use ResourceAction to support deletes, edits etc all at once, and previews for all
            if not yes:
                if input("Do delete? [Y/N] ").lower().strip() not in ["y"]:
                    print("Cancelling...")
                    return
            error = self._destroy()
            if error is not None:
                if debug:
                    traceback.print_exception(error, tb=error.__traceback__, value=error)
                else:
                    print(error)
        elif action is StackAction.DIFF:
            self._get_diff_config(output_format=diff_format)
            return
        else:
            raise ValueError("Unknown action")

        if config_path is not None:
            with open(config_path, "w") as f:
                f.write(json.dumps(self._output))

    def get_resource_id(self, unique_id: str) -> str:
        return self._output[unique_id]["id"]



from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator, validator, ValidationError

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.messaging import MessagingIntegrationType

SEGMENT_ID_ALL_DATA = "ALLDATA"


class MonitorType(Enum):
    MODEL_ACTIVITY = "model_activity"
    DATA_DRIFT = "data_drift"
    PREDICTION_DRIFT = "prediction_drift"
    MISSING_VALUES = "missing_values"
    PERFORMANCE_DEGRADATION = "performance_degradation"
    MODEL_STALENESS = "model_staleness"
    METRIC_CHANGE = "metric_change"
    CUSTOM_METRIC_MONITOR = "custom_metric"
    VALUES_RANGE = "values_range"
    NEW_VALUES = "new_values"


class DetectionMethod(Enum):
    ANOMALY = "anomaly"
    PERCENTAGE = "percentage"
    ABSOLUTE = "absolute"
    COMPARED_TO_SEGMENT = "segment"
    COMPARED_TO_TRAINING = "training"


class SourceType(Enum):
    SERVING = "SERVING"
    TRAINING = "TRAINING"


class PeriodType(Enum):
    HOURS = "h"
    DAYS = "d"
    WEEKS = "w"
    MONTHS = "M"


class TimePeriod(BaseModel):
    count: int
    type: PeriodType

    def to_string(self) -> str:
        return f"{self.count}{self.type.value}"

    @classmethod
    def from_string(cls, s) -> "TimePeriod":
        c = s[:-1]
        t = PeriodType(s[-1])
        return cls(count=c, type=t)


# TODO: Rename stuff to match Python conventions
class FocalConfiguration(BaseModel):
    source: SourceType = SourceType.SERVING
    timePeriod: Optional[TimePeriod] = None  # Missing for staleness
    skipPeriod: Optional[TimePeriod] = None
    alignBinsWithBaseline: Optional[bool] = None  # Needed (as True) for drift/histogram

    @validator("timePeriod", pre=True)
    def _parse_time_period(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @validator("skipPeriod", pre=True)
    def _parse_skip_period(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @validator("source")
    def _validate_source(cls, value):
        if value is not SourceType.SERVING:
            raise ValueError("For focal, source must always be serving")
        return value

    @validator("alignBinsWithBaseline")
    def _validate_align_bins_with_baseline(cls, value):
        if value is not True:
            raise ValueError("If alignBinsWithBaseline appears, it must be True")
        return value

    def to_dict(self):
        result = {
            "source": self.source.value,
        }
        if self.timePeriod is not None:
            result["timePeriod"] = self.timePeriod.to_string()
        if self.skipPeriod is not None:
            result["skipPeriod"] = self.skipPeriod.to_string()
        if self.alignBinsWithBaseline is not None:
            result["alignBinsWithBaseline"] = self.alignBinsWithBaseline

        return result


class BaselineConfiguration(BaseModel):
    source: SourceType
    timePeriod: Optional[TimePeriod] = None  # Actual time to compare to
    skipPeriod: Optional[TimePeriod] = None  # By default, equal to Focal. Doesn't exist on training
    aggregationPeriod: Optional[
        TimePeriod
    ] = None  # Does't exist on drift/histogram # TODO: Also on new values and value range. Check WTF is it
    segmentValue: Optional[Union[str, int, float]] = None
    segmentGroupId: Optional[str] = None

    @validator("timePeriod", pre=True)
    def _parse_time_period(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @validator("skipPeriod", pre=True)
    def _parse_skip_period(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @validator("aggregationPeriod", pre=True)
    def _parse_aggregation_period(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @validator("skipPeriod", always=True)
    def _validate_skip_period(cls, value, values):
        if values["source"] is SourceType.TRAINING:
            if value is not None:
                # TODO: Asserting this fails because UI does stupid things -.-
                # raise ValueError("For training baseline, skipPeriod must not appear")
                return None
        # TODO: This is basically defaulted in the global monitor validator
        # elif values["source"] is SourceType.SERVING:
        #     if value is None:
        #         raise ValueError("For serving baseline, skipPeriod must appear")
        return value

    @validator("timePeriod", always=True)
    def _validate_time_period(cls, value, values):
        if values["source"] is SourceType.SERVING:
            if value is None:
                raise ValueError("For serving baseline, timePeriod must appear")
        else:
            if value is not None:
                # TODO: Asserting this fails because UI does stupid things -.-
                # raise ValueError("For training baseline, timePeriod must not appear")
                return None
        return value

    def to_dict(self):
        result = {"source": self.source.value}
        if self.timePeriod is not None:
            result["timePeriod"] = self.timePeriod.to_string()
        if self.skipPeriod is not None:
            result["skipPeriod"] = self.skipPeriod.to_string()
        if self.aggregationPeriod is not None:
            result["aggregationPeriod"] = self.aggregationPeriod.to_string()
        # TODO: Validate these fields
        if self.segmentGroupId is not None:
            if self.segmentGroupId == SEGMENT_ID_ALL_DATA:
                result["segmentGroupId"] = None
            else:
                result["segmentGroupId"] = self.segmentGroupId
        if self.segmentValue is not None:
            result["segmentValue"] = self.segmentValue

        return result


class MetricType(Enum):
    COUNT = "count"
    MISSING_RATIO = "missing_ratio"
    LAST_VERSION_CREATION = "last_version_creation"
    HISTOGRAM = "histogram"
    MIN_MAX = "min_max"
    UNIQUE_VALUES = "unique_values"
    CUSTOM_METRIC = "custom_metric"
    MEAN = "mean"
    MEAN_LENGTH = "mean_length"
    MIN = "min"
    MIN_LENGTH = "min_length"
    MAX = "max"
    MAX_LENGTH = "max_length"
    SUM = "sum"
    MISSING_COUNT = "missing_count"
    MISSING = "missing"
    TP_COUNT = "tp_count"
    TN_COUNT = "tn_count"
    FP_COUNT = "fp_count"
    FN_COUNT = "fn_count"
    ACCURACY = "accuracy"
    RECALL = "recall"
    PRECISION = "precision"
    F1 = "f1"
    MSE = "mse"
    RMSE = "rmse"
    MAE = "mae"
    ACCURACY_AT_K = "accuracy_at_k"
    PRECISION_AT_K = "precision_at_k"
    RECALL_AT_K = "recall_at_k"
    MAP_AT_K = "map_at_k"
    MRR_AT_K = "mrr_at_k"
    NDCG_AT_K = "ndcg_at_k"
    JS_DISTANCE = "js_distance"
    KS_DISTANCE = "ks_distance"
    HELLINGER_DISTANCE = "hellinger_distance"
    EUCLIDEAN_DISTANCE = "euclidean_distance"
    PSI = "psi"
    ACCURACY_PER_CLASS = "accuracy_per_class"
    PRECISION_PER_CLASS = "precision_per_class"
    RECALL_PER_CLASS = "recall_per_class"
    F1_PER_CLASS = "f1_per_class"
    TP_COUNT_PER_CLASS = "tp_count_per_class"
    TN_COUNT_PER_CLASS = "tn_count_per_class"
    FP_COUNT_PER_CLASS = "fp_count_per_class"
    FN_COUNT_PER_CLASS = "fn_count_per_class"


class AverageMethod(Enum):
    MICRO = "micro"
    MACRO = "macro"
    WEIGHTED = "weighted"


class MetricConfiguration(BaseModel):
    type: MetricType
    # Necessary for custom metrics
    id: Optional[str] = None
    # Necessary for general metrics
    metricAtK: Optional[int] = None  # TODO: Almost always 3 -.- Gubkin
    threshold: Optional[float] = None
    metricPerClass: Optional[Any] = None
    average: Optional[AverageMethod] = None
    # TODO: Maybe add class for Aporia metrics

    # TODO: Add validators for params by metric types

    @validator("id", always=True)
    def _validate_id(cls, value, values):
        if values["type"] is MetricType.CUSTOM_METRIC:
            if value is None:
                raise ValueError("For custom metrics, id must appear")
        else:
            if value is not None:
                raise ValueError("id must only appear for custom metrics")
        return value

    def to_dict(self):
        result = {"type": self.type.value}
        if self.id is not None:
            result["id"] = self.id
        if self.metricAtK is not None:
            result["metricAtK"] = self.metricAtK
        if self.threshold is not None:
            result["threshold"] = self.threshold
        if self.metricPerClass is not None:
            result["metricPerClass"] = self.metricPerClass
        if self.average is not None:
            result["average"] = self.average.value

        return result


class LogicEvaluationType(Enum):
    RATIO = "RATIO"
    RANGE = "RANGE"
    TIME_SERIES_ANOMALY = "TIME_SERIES_ANOMALY"
    APORIA_DRIFT_SCORE = "APORIA_DRIFT_SCORE"
    MODEL_STALENESS = "MODEL_STALENESS"
    VALUES_RANGE = "VALUES_RANGE"


class ThresholdConfiguration(BaseModel):
    numeric: Optional[float] = None
    categorical: Optional[float] = None

    def to_dict(self):
        result = {}

        if self.numeric is not None:
            result["numeric"] = self.numeric
        if self.categorical is not None:
            result["categorical"] = self.categorical

        return result


class LogicEvaluationConfiguration(BaseModel):
    name: LogicEvaluationType
    # Needed for range (Both must exist, if one isn't used, it should be null)
    min: Optional[float] = None
    # Also needed for ratio (float) and staleness (TimePeriod)
    max: Optional[Union[float, TimePeriod]] = None
    # Needed for timeseries anomaly
    sensitivity: Optional[float] = None
    testOnlyIncrease: Optional[bool] = None
    # Needed for Aporia drift score
    thresholds: Optional[ThresholdConfiguration] = None
    # Needed for non-absolute Value range
    distance: Optional[float] = None
    # Needed for new-values value range
    new_values_count_threshold: Optional[int] = None
    new_values_ratio_threshold: Optional[float] = None

    @validator("max", pre=True)
    def _parse_max(cls, value):
        if isinstance(value, str):
            return TimePeriod.from_string(value)
        return value

    @root_validator()
    def _validate_logic_evaluation(cls, values):
        if values["name"] is LogicEvaluationType.RANGE:
            if isinstance(values["max"], TimePeriod):
                raise ValueError("max must be a float or None for Range logic evaluation")
            if (
                values["sensitivity"] is not None
                or values["testOnlyIncrease"] is not None
                or values["thresholds"] is not None
                or values["distance"] is not None
            ):
                raise ValueError("Only min and max can appear for Range logic evaluation")

        if values["name"] is LogicEvaluationType.VALUES_RANGE:
            if isinstance(values["max"], TimePeriod):
                raise ValueError("max must be a float or None for Range logic evaluation")
            if isinstance(values["min"], TimePeriod):
                raise ValueError("min must be a float or None for Range logic evaluation")
            if (
                values["sensitivity"] is not None
                or values["thresholds"] is not None
                or values["testOnlyIncrease"] is not None
            ):
                raise ValueError("Only min and max can appear for Value Range logic evaluation")

        if values["name"] is LogicEvaluationType.RATIO:
            if isinstance(values["max"], TimePeriod):
                raise ValueError("max must be a float for Ratio logic evaluation")
            if (
                values["sensitivity"] is not None
                or values["testOnlyIncrease"] is not None
                or values["thresholds"] is not None
                or values["distance"] is not None
                or values["new_values_count_threshold"] is not None
                or values["new_values_ratio_threshold"] is not None
            ):
                raise ValueError("Only min and max can appear for Ratio logic evaluation")

        if values["name"] is LogicEvaluationType.MODEL_STALENESS:
            if not isinstance(values["max"], TimePeriod):
                raise ValueError("max must be a TimePeriod for Stalenss logic evaluation")
            if (
                values["sensitivity"] is not None
                or values["testOnlyIncrease"] is not None
                or values["thresholds"] is not None
                or values["distance"] is not None
                or values["new_values_count_threshold"] is not None
                or values["new_values_ratio_threshold"] is not None
                or values["min"] is not None
            ):
                raise ValueError("Only max can appear for Staleness logic evaluation")

        if values["name"] is LogicEvaluationType.TIME_SERIES_ANOMALY:
            if values["sensitivity"] is None:
                raise ValueError("sensitivity must appear for Timeseries Anomaly Logic Evaluation")
            if (
                values["max"] is not None
                or values["min"] is not None
                or values["thresholds"] is not None
                or values["distance"] is not None
                or values["new_values_count_threshold"] is not None
                or values["new_values_ratio_threshold"] is not None
            ):
                raise ValueError(
                    "Only sensitivity can appear for Timeseries Anomaly logic evaluation"
                )

        if values["name"] is LogicEvaluationType.APORIA_DRIFT_SCORE:
            if values["thresholds"] is None:
                raise ValueError("thresholds must appear for Aporia Drift Score Logic Evaluation")
            if (
                values["max"] is not None
                or values["min"] is not None
                or values["sensitivity"] is not None
                or values["testOnlyIncrease"] is not None
                or values["distance"] is not None
                or values["new_values_count_threshold"] is not None
                or values["new_values_ratio_threshold"] is not None
            ):
                raise ValueError(
                    "Only thresholds can appear for Aporia Drift Score logic evaluation"
                )

        return values

    def to_dict(self):
        result = {"name": self.name.value}

        if self.min is not None:
            result["min"] = self.min
        if self.max is not None:
            if isinstance(self.max, TimePeriod):
                result["max"] = self.max.to_string()
            else:
                result["max"] = self.max
        if self.sensitivity is not None:
            result["sensitivity"] = self.sensitivity
        if self.testOnlyIncrease is not None:
            result["testOnlyIncrease"] = self.testOnlyIncrease
        if self.thresholds is not None:
            result["thresholds"] = self.thresholds.to_dict()

        if self.distance is not None:
            result["distance"] = self.distance

        if self.new_values_count_threshold is not None:
            result["new_values_count_threshold"] = self.new_values_count_threshold
        if self.new_values_ratio_threshold is not None:
            result["new_values_ratio_threshold"] = self.new_values_ratio_threshold

        return result


class PreConditionType(Enum):
    MIN_BASELINE_DATA_POINTS = "MIN_BASELINE_DATA_POINTS"
    MIN_FOCAL_DATA_POINTS = "MIN_FOCAL_DATA_POINTS"
    FOCAL_DATA_VALUE_IN_RANGE = "FOCAL_DATA_VALUE_IN_RANGE"


class PreConditionConfiguration(BaseModel):
    name: PreConditionType
    # Needed for MIN_BASELINE_DATA_POINTS
    value: Optional[int] = None
    # Needed for FOCAL_DATA_VALUE_IN_RANGE
    min: Optional[float] = None

    @root_validator()
    def _validate_pre_condition(cls, values):
        if values["name"] is PreConditionType.MIN_BASELINE_DATA_POINTS:
            if values["value"] is None:
                raise ValueError("value must appear for MIN_BASELINE_DATA_POINTS PreCondition")
            if values["min"] is not None:
                raise ValueError("Only value can appear for MIN_BASELINE_DATA_POINTS PreCondition")
        if values["name"] is PreConditionType.FOCAL_DATA_VALUE_IN_RANGE:
            # TODO: Min doesn't have to appear, appearantly. It just skips the test :(
            # if values["min"] is None:
            #     raise ValueError("min must appear for FOCAL_DATA_VALUE_IN_RANGE PreCondition")
            if values["value"] is not None:
                raise ValueError("Only min can appear for FOCAL_DATA_VALUE_IN_RANGE PreCondition")

        return values

    def to_dict(self):
        result = {"name": self.name.value}

        if self.min is not None:
            result["min"] = self.min
        if self.value is not None:
            result["value"] = self.value

        return result


class ActionType(Enum):
    ALERT = "ALERT"


class Severity(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class AlertType(Enum):
    MODEL_STALENESS = "model_staleness"
    METRIC_CHANGE = "metric_change"
    NEW_VALUES = "new_values"
    VALUES_RANGE = "values_range"
    DATA_DRIFT_ANOMALY = "data_drift_anomaly"
    DATA_DRIFT_SEGMENT_CHANGE = "data_drift_segment_change"
    DATA_DRIFT_TRAINING = "data_drift_training"
    PREDICTION_DRIFT_ANOMALY = "prediction_drift_anomaly"
    PREDICTION_DRIFT_SEGMENT_CHANGE = "prediction_drift_segment_change"
    PREDICTION_DRIFT_TRAINING = "prediction_drift_training"
    FEATURE_MISSING_VALUES_CHANGE = "feature_missing_values_change"
    FEATURE_MISSING_VALUES_ANOMALY = "feature_missing_values_anomaly"
    FEATURE_MISSING_VALUES_THRESHOLD = "feature_missing_values_threshold"
    FEATURE_MISSING_VALUES_SEGMENT_CHANGE = "feature_missing_values_segment_change"
    MODEL_ACTIVITY_THRESHOLD = "model_activity_threshold"
    MODEL_ACTIVITY_ANOMALY = "model_activity_anomaly"
    MODEL_ACTIVITY_CHANGE = "model_activity_change"


class VisualizationType(Enum):
    DISTRIBUTION_COMPARE_CHART = "distribution_compare_chart"
    RANGE_LINE_CHART = "range_line_chart"
    VALUE_OVER_TIME = "value_over_time"
    EMBEDDING_DRIFT_CHART = "embedding_drift_chart"
    VALUES_CANDLESTICK_CHART = "values_candlestick_chart"


class ActionConfiguration(BaseModel):
    type: ActionType
    severity: Severity
    alertType: AlertType
    description: str
    notification: List  # TODO: Look deeper into that
    action_schema: str = "v1"  # TODO: Realize what to do with it
    maxAlertsPerDay: Optional[int] = None
    visualization: Optional[VisualizationType] = None

    def to_dict(self):
        result = {
            "type": self.type.value,
            "severity": self.severity.value,
            "alertType": self.alertType.value,
            "description": self.description,
            "notification": self.notification,
            "schema": self.action_schema,
        }

        if self.maxAlertsPerDay is not None:
            result["maxAlertsPerDay"] = self.maxAlertsPerDay
        if self.visualization is not None:
            result["visualization"] = self.visualization.value

        return result


class ModelIdentification(BaseModel):
    id: str
    version: Optional[
        str
    ] = None  # None indicates all versions # TODO: Saw one saying "all_versions" as well?? Validate that

    def to_dict(self):
        return self.dict()


class SegmentIdentification(BaseModel):
    group: Optional[str] = None
    value: Optional[Union[str, int, float]] = None  # None indicates all segment values

    def to_dict(self):
        return self.dict()


class Identification(BaseModel):
    models: ModelIdentification
    segment: SegmentIdentification
    # TODO: Check if actuals are needed
    raw_inputs: Optional[List[str]] = None
    features: Optional[List[str]] = None
    predictions: Optional[List[str]] = None

    def to_dict(self):
        result = {
            "models": self.models.to_dict(),
            "segment": self.segment.to_dict(),
        }
        if self.raw_inputs is not None:
            result["raw_inputs"] = self.raw_inputs
        if self.features is not None:
            result["features"] = self.features
        if self.predictions is not None:
            result["predictions"] = self.predictions

        return result


class MonitorConfiguration(BaseModel):
    identification: Identification
    focal: FocalConfiguration
    metric: MetricConfiguration
    actions: List[ActionConfiguration]
    baseline: Optional[BaselineConfiguration] = None
    logicEvaluations: Optional[List[LogicEvaluationConfiguration]] = None
    preConditions: Optional[List[PreConditionConfiguration]] = None
    # TODO: Add cross-configuration validators

    @root_validator()
    def _validate_monitor_configuration(cls, values):
        if values.get("baseline") is not None:
            # For non-training monitors, skipPeriod defaults to the focal timePeriod, unless it's per segment
            if values["baseline"].source is SourceType.SERVING:
                # TODO: Fix that, as this sets skipPeriod even when using segments
                # TODO: CRITICAL BUG! This ruins the type infering
                if (
                    values["baseline"].segmentValue is None
                    and values["baseline"].segmentGroupId is None
                ):
                    if values["baseline"].skipPeriod is None:
                        values["baseline"].skipPeriod = values["focal"].timePeriod

            # For anomaly detection monitors, aggregationPeriod defaults to timePeriod
            if values["logicEvaluations"][0].name is LogicEvaluationType.TIME_SERIES_ANOMALY:
                if values["baseline"].aggregationPeriod is None:
                    values["baseline"].aggregationPeriod = values["focal"].timePeriod

        return values

    def to_dict(self):
        result = {
            "identification": self.identification.to_dict(),
            "configuration": {
                "focal": self.focal.to_dict(),
                "metric": self.metric.to_dict(),
                "actions": [action.to_dict() for action in self.actions],
            },
        }
        if self.baseline is not None:
            result["configuration"]["baseline"] = self.baseline.to_dict()
        if self.logicEvaluations is not None:
            result["configuration"]["logicEvaluations"] = [
                logic_evaluation.to_dict() for logic_evaluation in self.logicEvaluations
            ]
        if self.preConditions is not None:
            result["configuration"]["preConditions"] = [
                precondition.to_dict() for precondition in self.preConditions
            ]

        return result


class Monitor(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.type = MonitorType(data["type"])

    @classmethod
    def get_all(cls, client: Client, model_id: Optional[str] = None) -> List["Monitor"]:
        response = client.send_request("/monitors", "GET")

        client.assert_response(response)

        return [
            cls(client=client, data=entry)
            for entry in response.json()
            if model_id is None
            or entry["configuration"]["identification"]["models"]["id"] == model_id
        ]

    @classmethod
    def create(
        cls,
        client: Client,
        model_id: str,
        name: str,
        monitor_type: MonitorType,
        configuration: MonitorConfiguration,
        scheduling: Optional[str] = None,  # TODO: Optionally infer from configuration
        comment: Optional[str] = None,
        creator: Optional[str] = None,
        is_active: bool = True,
    ) -> "Monitor":
        monitor_type = MonitorType(monitor_type)

        if scheduling is None:
            if configuration.focal.timePeriod is None:
                scheduling = "0 * * * *"
            else:
                # if configuration.focal.timePeriod.count == 1:
                if configuration.focal.timePeriod.type is PeriodType.HOURS:
                    scheduling = "*/5 * * * *"  # Every 5 minutes
                elif configuration.focal.timePeriod.type is PeriodType.DAYS:
                    scheduling = "0 */4 * * *"  # Every 4 hours
                elif configuration.focal.timePeriod.type is PeriodType.WEEKS:
                    scheduling = "0 */12 * * *"  # Every 12 hours
                elif configuration.focal.timePeriod.type is PeriodType.MONTHS:
                    scheduling = "0 0 */2 * *"

        configuration.identification.models.id = model_id

        if scheduling is None:
            raise ValueError("Scheduling must be defined")

        response = client.send_request(
            "/monitors",
            "POST",
            {
                "name": name,
                "type": monitor_type.value,
                "scheduling": scheduling,
                "configuration": configuration.to_dict(),
                "comment": comment,
                "creator_id": creator,
                "is_active": is_active,
            },
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Monitor":
        response = client.send_request(f"/monitors/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(
        self,
        name: Optional[str] = None,
        monitor_type: Optional[MonitorType] = None,
        scheduling: Optional[str] = None,
        configuration: Optional[MonitorConfiguration] = None,
        comment: Optional[str] = None,
        creator: Optional[str] = None,
        is_active: bool = True,
        **kwargs,
    ):
        args = {}
        if name is not None:
            args["name"] = name
        if monitor_type is not None:
            args["type"] = MonitorType(monitor_type).value
        if scheduling is not None:
            args["scheduling"] = scheduling
        if configuration is not None:
            args["configuration"] = configuration.to_dict()
        if comment is not None:
            args["comment"] = comment
        if creator is not None:
            args["creator"] = creator
        if is_active is not None:
            args["is_active"] = is_active
        # response = self.client.send_request(f"/monitors/{self.id}", "PUT", args)
        response = self.client.send_request(
            f"/monitors/{self.id}",
            "PUT",
            args,
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/monitors/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/monitors/{id}", "DELETE")
        client.assert_response(response)


def __get_alert_description_range(min: Any, max: Any) -> str:
    if min is not None and max is not None:
        return "between <b>{min_threshold}</b> to <b>{max_threshold}</b>"
    if min is not None:
        return "above <b>{min_threshold}</b>"
    if max is not None:
        return "below <b>{max_threshold}</b>"
    raise ValueError("Bad configuration: Both min and max weren't supplied")


def __get_alert_field_description_if_needed(metric: MetricType) -> str:
    metrics_with_fields = [
        "missing_count",
        "mean",
        "min",
        "max",
        "sum",
        "variance",
        "std",
        "mse",
        "rmse",
        "mae",
        "precision",
        "recall",
        "f1",
        "accuracy",
        "tp_count",
        "tn_count",
        "fp_count",
        "fn_count",
        "wape",
        "mape",
        "auc_roc",
        "logloss",
    ]
    if metric.value in metrics_with_fields:
        return "of <b>{field}</b> {importance} "
    return ""


def _MonitorParamBuilder(
    alert_type: AlertType,
    metric_builder: Callable[..., MetricConfiguration],
    description_builder: Callable[..., str],
    logic_evaluation_builder: Callable[..., LogicEvaluationConfiguration],
    precondition_builder: Optional[Callable[..., Optional[PreConditionConfiguration]]] = None,
    needed_args: Optional[List[Union[str, List[str]]]] = None,
    visualization_builder: Optional[Callable[..., VisualizationType]] = None,
):
    class _MonitorParams:
        def __init__(self, **kw):
            self.alert_type = AlertType(alert_type)
            if needed_args is not None:
                for arg in needed_args:
                    if isinstance(arg, list):
                        if not any(optional_arg in kw.keys() for optional_arg in arg):
                            raise ValueError(
                                f"One of {arg} must be supplied for {alert_type.value} monitor"
                            )
                    else:
                        if kw[arg] is None:
                            raise ValueError(
                                f"Argument {arg} must be supplied for {alert_type.value} monitor"
                            )
            self.kw = kw

        @property
        def metric(self):
            return metric_builder(**self.kw)

        @property
        def description(self):
            return description_builder(**self.kw)

        @property
        def logic_evaluation(self):
            return logic_evaluation_builder(**self.kw)

        @property
        def precondition(self):
            if precondition_builder is not None:
                return precondition_builder(**self.kw)
            return None

        @property
        def visualization(self):
            if visualization_builder is not None:
                return VisualizationType(visualization_builder(**self.kw))
            return None

    return _MonitorParams


_MONITOR_PARAMETERS = {
    MonitorType.MODEL_ACTIVITY: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.COUNT),
            alert_type=AlertType.MODEL_ACTIVITY_ANOMALY,
            visualization_builder=lambda **kw: VisualizationType.RANGE_LINE_CHART,
            description_builder=lambda **kw: "An anomaly in the number of total predictions was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on total predictions count history, the count was expected to be between <b>{last_lower_bound}</b> to <b>{last_upper_bound}</b>, but <b>{focal_value}</b> was received. <br /><br /> Total prediction anomaly might occur because: <ul><li>Natural changes in model invocations</li><li>Serving environment fault</li><li>Malicious attempt to analyse model behaviour</li></ul>",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.TIME_SERIES_ANOMALY,
                sensitivity=kw["sensitivity"],
                testOnlyIncrease=kw["testOnlyIncrease"],
            ),
            needed_args=["sensitivity", "baseline"],
        ),
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.COUNT),
            alert_type=AlertType.MODEL_ACTIVITY_CHANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            description_builder=lambda **kw: (
                "An anomaly in the number of total predictions was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on total predictions count history average and on defined ratio threshold, the count was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received. <br /><br /> Total prediction anomaly might occur because: <ul><li>Natural changes in model invocations</li><li>Serving environment fault</li><li>Malicious attempt to analyse model behaviour</li></ul>"
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["percentage", "baseline"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.COUNT),
            alert_type=AlertType.MODEL_ACTIVITY_THRESHOLD,
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            description_builder=lambda **kw: (
                "An anomaly in the number of total predictions within the defined limits was detected.<br />The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>.<br /><br />Based on defined limits, the count was expected to be "
                + __get_alert_description_range(min=kw["min"], max=kw["max"])
                + ", but <b>{focal_value}</b> was received.<br />"
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=kw["min"], max=kw["max"]
            ),
            needed_args=[["min", "max"]],
        ),
    },
    MonitorType.DATA_DRIFT: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            alert_type=AlertType.DATA_DRIFT_ANOMALY,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A data drift was detected in feature <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>last {{baseline_time_period}} ({{baseline_times}})</b>. <br /><br /> Data drift can have a significant effect on model behavior and may lead to unexpected results.<br /><br /> Data drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            alert_type=AlertType.DATA_DRIFT_SEGMENT_CHANGE,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A data drift was detected in feature <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>last {{baseline_time_period}} ({{baseline_times}})</b> <b>{{baseline_segment}}</b>. <br /><br /> Data drift can have a significant effect on model behavior and may lead to unexpected results.<br /><br /> Data drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            alert_type=AlertType.DATA_DRIFT_TRAINING,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A data drift was detected in feature <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>Training</b> data. <br /><br /> Data drift can have a significant effect on model behavior and may lead to unexpected results.<br /><br /> Data drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
    },
    MonitorType.PREDICTION_DRIFT: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            alert_type=AlertType.PREDICTION_DRIFT_ANOMALY,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A prediction drift was detected in prediction <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>last {{baseline_time_period}} ({{baseline_times}})</b>. <br /><br /> Prediction drift indicates a significant change in model's behavior. In some cases, it is a strong indicator for concept drift.<br /><br /> Prediction drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            alert_type=AlertType.PREDICTION_DRIFT_SEGMENT_CHANGE,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A prediction drift was detected in prediction <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>last {{baseline_time_period}} ({{baseline_times}})</b> <b>{{baseline_segment}}</b>. <br /><br /> Prediction drift indicates a significant change in model's behavior. In some cases, it is a strong indicator for concept drift.<br /><br /> Prediction drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            alert_type=AlertType.PREDICTION_DRIFT_TRAINING,
            visualization_builder=lambda **kw: VisualizationType.EMBEDDING_DRIFT_CHART
            if kw["is_embedding"]
            else VisualizationType.DISTRIBUTION_COMPARE_CHART,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.EUCLIDEAN_DISTANCE if kw["is_embedding"] else MetricType.HISTOGRAM
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=0, max=kw["max"]
            )
            if kw["is_embedding"]
            else LogicEvaluationConfiguration(
                name=LogicEvaluationType.APORIA_DRIFT_SCORE, thresholds=kw["thresholds"]
            ),
            description_builder=lambda **kw: f"A prediction drift was detected in prediction <b>'{{field}}'</b> {{importance}}.{'' if kw['is_embedding'] else '{drift_score_text}'}<br /> The drift was observed in the <b>{{model}}</b> model, in version <b>{{model_version}}</b> for the <b>last {{focal_time_period}} ({{focal_times}})</b> <b>{{focal_segment}}</b> compared to the <b>Training</b> data. <br /><br /> Prediction drift indicates a significant change in model's behavior. In some cases, it is a strong indicator for concept drift.<br /><br /> Prediction drift might occur because: <ul><li>Natural changes in data</li><li>Data store / provider schema changes</li><li>Data store / provider issues</li><li>Data processing issues</li></ul>",
            needed_args=[["thresholds", "max"], "baseline"],
        ),
    },
    MonitorType.MISSING_VALUES: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MISSING_RATIO),
            alert_type=AlertType.FEATURE_MISSING_VALUES_ANOMALY,
            visualization_builder=lambda **kw: VisualizationType.RANGE_LINE_CHART,
            description_builder=lambda **kw: "An anomaly was detected in the ratio of missing values of feature <b>'{field}'</b> {importance}.<br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on missing ratio history, the ratio was expected to be between <b>{last_lower_bound}</b> to <b>{last_upper_bound}</b>, but <b>{focal_value}</b> was received. <br /><br /> Missing data can have a significant effect on model behavior and may lead to unexpected results. <br /><br /> Missing data might occur because: <ul><li>Serving environment fault</li><li>Data store / provider schema changes</li><li>Changes in internal API</li><li>Changes in model subject input</li></ul>",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.TIME_SERIES_ANOMALY,
                sensitivity=kw["sensitivity"],
                testOnlyIncrease=kw["testOnlyIncrease"],
            ),
            precondition_builder=lambda **kw: (
                PreConditionConfiguration(
                    name=PreConditionType.FOCAL_DATA_VALUE_IN_RANGE, min=kw["min"]
                )
                if kw["min"] is not None
                else None
            ),
            needed_args=["sensitivity", "baseline"],
        ),
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MISSING_RATIO),
            alert_type=AlertType.FEATURE_MISSING_VALUES_CHANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            description_builder=lambda **kw: (
                "An anomaly in the ratio of missing values of feature <b>'{field}'</b> {importance} was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on missing ratio history average and on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received. <br /><br /> Missing data can have a significant effect on model behavior and may lead to unexpected results. <br /><br /> Missing data might occur because: <ul><li>Serving environment fault</li><li>Data store / provider schema changes</li><li>Changes in internal API</li><li>Changes in model subject input</li></ul>"
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            precondition_builder=lambda **kw: (
                PreConditionConfiguration(
                    name=PreConditionType.FOCAL_DATA_VALUE_IN_RANGE, min=kw["min"]
                )
                if kw["min"] is not None
                else None
            ),
            needed_args=["percentage", "baseline"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MISSING_RATIO),
            alert_type=AlertType.FEATURE_MISSING_VALUES_THRESHOLD,
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            description_builder=lambda **kw: (
                "An anomaly in the ratio of missing values of feature <b>'{field}'</b> {importance} within the defined limits was detected.<br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on defined limits, the missing ratio was expected to be "
                + __get_alert_description_range(min=kw["min"], max=kw["max"])
                + ", but <b>{focal_value}</b> was received.<br /><br /> Missing data can have a significant effect on model behavior and may lead to unexpected results. <br /><br /> Missing data might occur because: <ul><li>Serving environment fault</li><li>Data store / provider schema changes</li><li>Changes in internal API</li><li>Changes in model subject input</li></ul>"
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=kw["min"], max=kw["max"]
            ),
            needed_args=[["min", "max"]],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MISSING_RATIO),
            alert_type=AlertType.FEATURE_MISSING_VALUES_SEGMENT_CHANGE,
            visualization_builder=lambda **kw: VisualizationType.DISTRIBUTION_COMPARE_CHART,
            description_builder=lambda **kw: (
                "An anomaly in the ratio of missing values of feature <b>'{field}'</b> {importance} within the defined limits was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b> <b>{baseline_segment}</b>. <br /><br /> Based on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 + (kw["percentage"] / 100), max=1 - (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received. <br /><br /> Missing data can have a significant effect on model behavior and may lead to unexpected results. <br /><br /> Missing data might occur because: <ul><li>Serving environment fault</li><li>Data store / provider schema changes</li><li>Changes in internal API</li><li>Changes in model subject input</li></ul>"
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                max=1 + (kw["percentage"] / 100),
                min=1 - (kw["percentage"] / 100),
            ),
            precondition_builder=lambda **kw: (
                PreConditionConfiguration(
                    name=PreConditionType.FOCAL_DATA_VALUE_IN_RANGE, min=kw["min"]
                )
                if kw["min"] is not None
                else None
            ),
            needed_args=["percentage", "baseline"],
        ),
    },
    MonitorType.PERFORMANCE_DEGRADATION: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.RANGE_LINE_CHART,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history, the ratio was expected to be "
                + "between <b>{last_lower_bound}</b> to <b>{last_upper_bound}</b>, but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.TIME_SERIES_ANOMALY,
                sensitivity=kw["sensitivity"],
                testOnlyIncrease=kw["testOnlyIncrease"],
            ),
            needed_args=["metric", "sensitivity", "baseline"],
        ),
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history average and on defined threshold, the value was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "percentage", "baseline"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected.<br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on defined limits, the metric value was expected to be "
                + __get_alert_description_range(min=kw["min"], max=kw["max"])
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE,
                min=kw["min"],
                max=kw["max"],
            ),
            needed_args=["metric", ["min", "max"]],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b> <b>{baseline_segment}</b>. <br /><br /> Based on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "percentage", "baseline"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to <b>Training</b> data. <br /><br /> Based on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "percentage", "baseline"],
        ),
    },
    MonitorType.METRIC_CHANGE: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.RANGE_LINE_CHART,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history, the ratio was expected to be "
                + "between <b>{last_lower_bound}</b> to <b>{last_upper_bound}</b>, but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.TIME_SERIES_ANOMALY,
                sensitivity=kw["sensitivity"],
                testOnlyIncrease=kw["testOnlyIncrease"],
            ),
            needed_args=["metric", "baseline", "sensitivity"],
        ),
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history average and on defined threshold, the value was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "baseline", "percentage"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected.<br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on defined limits, the metric value was expected to be "
                + __get_alert_description_range(min=kw["min"], max=kw["max"])
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=kw["min"], max=kw["max"]
            ),
            needed_args=["metric", ["min", "max"]],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b> <b>{baseline_segment}</b>. <br /><br /> Based on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "baseline", "percentage"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=kw["metric"],
                metricAtK=kw["k"],
                threshold=kw["threshold"],
                metricPerClass=kw["prediction_class"],
                average=kw["average_method"],
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType(kw["metric"]))
                + "within the defined limits was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to <b>Training</b> data. <br /><br /> Based on defined ratio threshold, the ratio was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["metric", "baseline", "percentage"],
        ),
    },
    MonitorType.MODEL_STALENESS: {
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            alert_type=AlertType.MODEL_STALENESS,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.LAST_VERSION_CREATION),
            description_builder=lambda **kw: "Model <b>{model}</b> was detected as stale. <br /> The current version in production <b>{model_version}</b> was deployed over <b>{time_threshold}</b> ago.<br /><br /> It is recommended to update the model in <b>{environment}</b>.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.MODEL_STALENESS, max=kw["staleness_period"]
            ),
            needed_args=["staleness_period"],
        ),
    },
    MonitorType.CUSTOM_METRIC_MONITOR: {
        DetectionMethod.ANOMALY: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.RANGE_LINE_CHART,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.CUSTOM_METRIC, id=kw["custom_metric_id"]
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType.CUSTOM_METRIC)
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history, the ratio was expected to be "
                + "between <b>{last_lower_bound}</b> to <b>{last_upper_bound}</b>, but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.TIME_SERIES_ANOMALY,
                sensitivity=kw["sensitivity"],
                testOnlyIncrease=kw["testOnlyIncrease"],
            ),
            needed_args=["custom_metric_id", "baseline", "sensitivity"],
        ),
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.CUSTOM_METRIC, id=kw["custom_metric_id"]
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType.CUSTOM_METRIC)
                + "was detected. <br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on metric history average and on defined threshold, the value was expected to be "
                + __get_alert_description_range(
                    min=1 - (kw["percentage"] / 100), max=1 + (kw["percentage"] / 100)
                )
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RATIO,
                min=1 - (kw["percentage"] / 100),
                max=1 + (kw["percentage"] / 100),
            ),
            needed_args=["custom_metric_id", "baseline", "percentage"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            visualization_builder=lambda **kw: VisualizationType.VALUE_OVER_TIME,
            alert_type=AlertType.METRIC_CHANGE,
            metric_builder=lambda **kw: MetricConfiguration(
                type=MetricType.CUSTOM_METRIC, id=kw["custom_metric_id"]
            ),
            description_builder=lambda **kw: (
                "An anomaly in the value of the <b>'{metric}'</b> "
                + __get_alert_field_description_if_needed(MetricType.CUSTOM_METRIC)
                + "within the defined limits was detected.<br /> The anomaly was observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on defined limits, the metric value was expected to be "
                + __get_alert_description_range(min=kw["min"], max=kw["max"])
                + ", but <b>{focal_value}</b> was received."
            ),
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.RANGE, min=kw["min"], max=kw["max"]
            ),
            needed_args=["custom_metric_id", ["min", "max"]],
        ),
    },
    MonitorType.VALUES_RANGE: {
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            alert_type=AlertType.VALUES_RANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUES_CANDLESTICK_CHART,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MIN_MAX),
            description_builder=lambda **kw: "Unexpected values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b>. <br /><br /> Based on defined limits, the values were expected to be <b>{value_thresholds}</b>, but values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE, distance=kw["distance"]
            ),
            needed_args=["distance", "baseline"],
        ),
        DetectionMethod.ABSOLUTE: _MonitorParamBuilder(
            alert_type=AlertType.VALUES_RANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUES_CANDLESTICK_CHART,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MIN_MAX),
            description_builder=lambda **kw: "Unexpected values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b>. <br /><br /> Based on defined limits, the values were expected to be <b>{value_thresholds}</b>, but values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE, min=kw["min"], max=kw["max"]
            ),
            needed_args=[["min", "max"]],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            alert_type=AlertType.VALUES_RANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUES_CANDLESTICK_CHART,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MIN_MAX),
            description_builder=lambda **kw: "Unexpected values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b> <b>{baseline_segment}</b>. <br /><br /> Based on defined limits, the values were expected to be <b>{value_thresholds}</b>, but values <b>{unexpected_values}</b> were received.'",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE, distance=kw["distance"]
            ),
            needed_args=["distance", "baseline"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            alert_type=AlertType.VALUES_RANGE,
            visualization_builder=lambda **kw: VisualizationType.VALUES_CANDLESTICK_CHART,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.MIN_MAX),
            description_builder=lambda **kw: "Unexpected values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to <b>Training</b> data. <br /><br /> Based on defined limits, the values were expected to be <b>{value_thresholds}</b>, but values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE, distance=kw["distance"]
            ),
            needed_args=["distance", "baseline"],
        ),
    },
    MonitorType.NEW_VALUES: {
        DetectionMethod.PERCENTAGE: _MonitorParamBuilder(
            alert_type=AlertType.NEW_VALUES,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.UNIQUE_VALUES),
            description_builder=lambda **kw: "New values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b>. <br /><br /> Based on defined limits, the following new values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE,
                new_values_ratio_threshold=kw["new_values_ratio_threshold"],
                new_values_count_threshold=kw["new_values_count_threshold"],
            ),
            needed_args=[["new_values_ratio_threshold", "new_values_count_threshold"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_SEGMENT: _MonitorParamBuilder(
            alert_type=AlertType.NEW_VALUES,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.UNIQUE_VALUES),
            description_builder=lambda **kw: "New values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to the <b>last {baseline_time_period} ({baseline_times})</b> <b>{baseline_segment}</b>. <br /><br /> Based on defined limits, the following new values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE,
                new_values_ratio_threshold=kw["new_values_ratio_threshold"],
                new_values_count_threshold=kw["new_values_count_threshold"],
            ),
            needed_args=[["new_values_ratio_threshold", "new_values_count_threshold"], "baseline"],
        ),
        DetectionMethod.COMPARED_TO_TRAINING: _MonitorParamBuilder(
            alert_type=AlertType.NEW_VALUES,
            metric_builder=lambda **kw: MetricConfiguration(type=MetricType.UNIQUE_VALUES),
            description_builder=lambda **kw: "New values were detected in feature <b>'{field}'</b> {importance}. <br /> The values were observed in the <b>{model}</b> model, in version <b>{model_version}</b> for the <b>last {focal_time_period} ({focal_times})</b> <b>{focal_segment}</b> compared to <b>Training</b> data. <br /><br /> Based on defined limits, the following new values <b>{unexpected_values}</b> were received.",
            logic_evaluation_builder=lambda **kw: LogicEvaluationConfiguration(
                name=LogicEvaluationType.VALUES_RANGE,
                new_values_ratio_threshold=kw["new_values_ratio_threshold"],
                new_values_count_threshold=kw["new_values_count_threshold"],
            ),
            needed_args=[["new_values_ratio_threshold", "new_values_count_threshold"], "baseline"],
        ),
    },
}


def create_monitor_configuration(
    monitor_type: MonitorType,
    detection_method: DetectionMethod,
    focal: FocalConfiguration,
    severity: Severity,
    baseline: Optional[BaselineConfiguration] = None,
    raw_inputs: Optional[List[str]] = None,
    features: Optional[List[str]] = None,
    predictions: Optional[List[str]] = None,
    is_embedding_monitor: bool = False,
    metric: Optional[Union[MetricType, str]] = None,
    percentage: Optional[int] = None,
    thresholds: Optional[ThresholdConfiguration] = None,
    sensitivity: Optional[float] = None,
    alert_on_increase_only: Optional[bool] = None,
    min: Optional[int] = None,
    max: Optional[int] = None,
    distance: Optional[float] = None,
    new_values_ratio_threshold: Optional[float] = None,
    new_values_count_threshold: Optional[int] = None,
    staleness_period: Optional[TimePeriod] = None,
    min_focal_prediction_count: Optional[int] = None,
    min_baseline_prediction_count: Optional[int] = None,
    version: Optional[str] = None,
    segment_identification: Optional[SegmentIdentification] = None,
    messaging: Optional[Dict[MessagingIntegrationType, List[str]]] = None,
    emails: Optional[List[str]] = None,
    max_alerts_per_day: Optional[int] = None,
    k: Optional[int] = None,
    prediction_class: Optional[Any] = None,
    prediction_threshold: Optional[float] = None,
    average_method: Optional[AverageMethod] = None,
    custom_metric_id: Optional[str] = None,
) -> MonitorConfiguration:
    """Creates a monitor configuration by the given parameters."""
    if segment_identification is None:
        segment_identification = SegmentIdentification()

    monitor_type = MonitorType(monitor_type)
    detection_method = DetectionMethod(detection_method)
    severity = Severity(severity)

    preconditions = []

    if min_baseline_prediction_count is not None:
        if baseline is None:
            raise ValueError("Can't specify min_baseline_prediction_count for monitors with no baseline")
        preconditions.append(
            PreConditionConfiguration(
                name=PreConditionType.MIN_BASELINE_DATA_POINTS, value=min_baseline_prediction_count
            )
        )
    if min_focal_prediction_count is not None:
        preconditions.append(
            PreConditionConfiguration(
                name=PreConditionType.MIN_FOCAL_DATA_POINTS, value=min_focal_prediction_count
            )
        )

    monitor_params = _MONITOR_PARAMETERS[monitor_type][detection_method](
        percentage=percentage,
        thresholds=thresholds,
        sensitivity=sensitivity,
        testOnlyIncrease=alert_on_increase_only,
        min=min,
        max=max,
        new_values_count_threshold=new_values_count_threshold,
        new_values_ratio_threshold=new_values_ratio_threshold,
        distance=distance,
        staleness_period=staleness_period,
        baseline=baseline,
        is_embedding=is_embedding_monitor,
        metric=metric,
        k=k,
        threshold=prediction_threshold,
        prediction_class=prediction_class,
        average_method=average_method,
        custom_metric_id=custom_metric_id,
    )

    notifications = []
    if emails is not None:
        notifications.append({"type": "EMAIL", "emails": emails})

    if messaging is not None:
        for integration_type, integration_ids in messaging.items():
            integration_type = MessagingIntegrationType(integration_type)
            for integration_id in integration_ids:
                notifications.append(
                    {"type": integration_type.value, "integration_id": integration_id}
                )

    if monitor_params.precondition is not None:
        preconditions.append(monitor_params.precondition)

    if monitor_params.metric.type in [MetricType.HISTOGRAM, MetricType.EUCLIDEAN_DISTANCE]:
        focal.alignBinsWithBaseline = True

    return MonitorConfiguration(
        identification=Identification(
            models=ModelIdentification(id="", version=version),
            segment=segment_identification,
            raw_inputs=raw_inputs,
            features=features,
            predictions=predictions,
        ),
        metric=monitor_params.metric,
        focal=focal,
        baseline=baseline,
        preConditions=preconditions,
        logicEvaluations=[monitor_params.logic_evaluation],
        actions=[
            ActionConfiguration(
                type=ActionType.ALERT,
                severity=severity,
                alertType=monitor_params.alert_type,
                description=monitor_params.description,
                notification=notifications,
                maxAlertsPerDay=max_alerts_per_day,
                visualization=monitor_params.visualization,
            )
        ],
    )


def parse_monitor_configuration(configuration: Dict) -> MonitorConfiguration:
    config = configuration["configuration"]
    identification = configuration["identification"]

    return MonitorConfiguration(
        identification=identification,
        focal=config.get("focal"),
        baseline=config.get("baseline"),
        actions=[ActionConfiguration(**action) for action in config["actions"]],
        metric=config["metric"],
        logicEvaluations=[
            LogicEvaluationConfiguration(**evaluation)
            for evaluation in config.get("logicEvaluations", [])
        ],
        preConditions=[
            PreConditionConfiguration(**precondition)
            for precondition in config.get("preConditions", [])
        ],
    )


def infer_detection_method(configuration: MonitorConfiguration) -> DetectionMethod:
    logic_evaluation = configuration.logicEvaluations[0]
    if logic_evaluation.name is LogicEvaluationType.MODEL_STALENESS:
        return DetectionMethod.ABSOLUTE
    if logic_evaluation.name is LogicEvaluationType.RANGE:
        if configuration.metric.type is not MetricType.EUCLIDEAN_DISTANCE:
            return DetectionMethod.ABSOLUTE
    if logic_evaluation.name is LogicEvaluationType.TIME_SERIES_ANOMALY:
        return DetectionMethod.ANOMALY
    if logic_evaluation.name is LogicEvaluationType.APORIA_DRIFT_SCORE or (
        logic_evaluation.name is LogicEvaluationType.RANGE
        and configuration.metric.type is MetricType.EUCLIDEAN_DISTANCE
    ):
        if configuration.actions[0].alertType in [
            AlertType.DATA_DRIFT_ANOMALY,
            AlertType.PREDICTION_DRIFT_ANOMALY,
        ]:
            return DetectionMethod.ANOMALY
        elif configuration.actions[0].alertType in [
            AlertType.DATA_DRIFT_SEGMENT_CHANGE,
            AlertType.PREDICTION_DRIFT_SEGMENT_CHANGE,
        ]:
            return DetectionMethod.COMPARED_TO_SEGMENT
        elif configuration.actions[0].alertType in [
            AlertType.DATA_DRIFT_TRAINING,
            AlertType.PREDICTION_DRIFT_TRAINING,
        ]:
            return DetectionMethod.COMPARED_TO_TRAINING
    if configuration.baseline is None:
        return DetectionMethod.ABSOLUTE
    if configuration.baseline.source is SourceType.TRAINING:
        return DetectionMethod.COMPARED_TO_TRAINING
    if configuration.baseline.skipPeriod is None:
        return DetectionMethod.COMPARED_TO_SEGMENT
    if configuration.baseline.segmentGroupId is not None:
        return DetectionMethod.COMPARED_TO_SEGMENT
    return DetectionMethod.PERCENTAGE



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Tuple

from aporia.sdk.client import Client


class NonDeletableResourceException(Exception):
    pass


class CompareStatus(Enum):
    SAME = "same"
    UPDATEABLE = "updateable"
    MISMATCHED = "mismatched"


class BaseResource(ABC):
    @abstractmethod
    def __init__(self, resource_name: str, /, **kwargs):
        ...

    @abstractmethod
    def compare(self, resource_data: Dict) -> CompareStatus:
        ...

    @abstractmethod
    def setarg(self, arg_name: str, arg_value: Any):
        ...

    @abstractmethod
    def create(self, client: Client) -> Tuple[str, Dict]:
        ...

    @abstractmethod
    def read(self, client: Client, id: str) -> Dict:
        ...

    @abstractmethod
    def update(self, client: Client, id: str) -> Dict:
        ...

    @abstractmethod
    def delete(self, client: Client, id: str):
        ...

    @abstractmethod
    def get_diff(self, resource_data: Dict) -> Dict:
        ...



from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from aporia.as_code.resources.base import BaseResource, CompareStatus
from aporia.as_code.resources.custom_metrics import CustomMetric
from aporia.as_code.resources.monitor import Monitor
from aporia.as_code.resources.segment import Segment
from aporia.as_code.resources.version import Version
from aporia.sdk.client import Client
from aporia.sdk.models import Model as _Model
from aporia.sdk.models import ModelColor, ModelIcon, ModelType, NoOwner, ModelAggregationPeriod


class Model(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        # Model args
        type: Union[ModelType, str],
        name: Optional[str] = None,
        description: Optional[str] = None,
        icon: Optional[ModelIcon] = None,
        color: Optional[ModelColor] = None,
        owner: Optional[Union[str, NoOwner]] = NoOwner(),
        aggregation_period: Optional[Union[ModelAggregationPeriod, str]] = None,
        # Model sub-resources
        versions: Optional[List[Version]] = None,
        segments: Optional[List[Segment]] = None,
        custom_metrics: Optional[List[CustomMetric]] = None,
        monitors: Optional[List[Monitor]] = None,
    ):
        self.sub_resources = []
        self.name = resource_name
        if name is None:
            name = resource_name

        type = ModelType(type)
        self._args = {"name": name, "model_type": type}

        if description is not None:
            self._args["description"] = description
        if icon is not None:
            icon = ModelIcon(icon)
            self._args["icon"] = icon
        if color is not None:
            color = ModelColor(color)
            self._args["color"] = color
        if not isinstance(owner, NoOwner):
            self._args["owner"] = owner
        if aggregation_period is not None:
            self._args["aggregation_period"] = ModelAggregationPeriod(aggregation_period)

        for version in versions or []:
            self.sub_resources.append(
                (version, lambda data, version: version.setarg("model_id", data["id"]))
            )
        for segment in segments or []:
            self.sub_resources.append(
                (segment, lambda data, segment: segment.setarg("model_id", data["id"]))
            )
        for custom_metric in custom_metrics or []:
            self.sub_resources.append(
                (
                    custom_metric,
                    lambda data, custom_metric: custom_metric.setarg("model_id", data["id"]),
                )
            )
        for monitor in monitors or []:
            self.sub_resources.append(
                (monitor, lambda data, monitor: monitor.setarg("model_id", data["id"]))
            )

    def compare(self, resource_data: Dict) -> CompareStatus:
        checks = [
            self._args["name"] == resource_data["name"],
            self._args["model_type"].value == resource_data["type"],
        ]
        for k in ["description", "icon", "color", "owner"]:
            if k in self._args:
                if isinstance(self._args[k], Enum):
                    checks.append(self._args[k].value == resource_data[k])
                else:
                    checks.append(self._args[k] == resource_data[k])
        if all(checks):
            return CompareStatus.SAME
        return CompareStatus.UPDATEABLE

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        model = _Model.create(client=client, **self._args)
        return model.id, model.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _Model.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        model = _Model.read(client=client, id=id)
        model.update(**self._args)
        return model.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        _Model.delete_by_id(client=client, id=id)

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        if self._args["name"] != resource_data["name"]:
            diffs["name"] = (resource_data["name"], self._args["name"])
        if self._args["model_type"].value != resource_data["type"]:
            diffs["model_type"] = (resource_data["type"], self._args["model_type"])
        for k in ["description", "icon", "color", "owner"]:
            if k in self._args:
                if isinstance(self._args[k], Enum):
                    if self._args[k].value != resource_data[k]:
                        diffs[k] = (resource_data[k], self._args[k].value)
                else:
                    if self._args[k] != resource_data[k]:
                        diffs[k] = (resource_data[k], self._args[k])
        return diffs



from enum import Enum
from typing import Dict, List, Optional

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client


class MessagingIntegrationType(Enum):
    SLACK = "SLACK"
    WEBHOOK = "WEBHOOK"
    TEAMS = "TEAMS"
    DATADOG = "DATADOG"


class MessagingIntegration(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.type = MessagingIntegrationType(data["type"])

    @classmethod
    def get_all(
        cls, client: Client, integration_type: Optional[MessagingIntegrationType] = None
    ) -> List["MessagingIntegration"]:
        response = client.send_request(
            f"/integrate/query{'' if integration_type is None else f'/{integration_type.value}'}",
            "GET",
            url_search_replace=("/api/v1/", "/v1/messaging-service/"),
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(cls, client: Client, name: str, model_id: str) -> "MessagingIntegration":
        raise NotImplementedError()

    @classmethod
    def read(cls, client: Client, id: str) -> "MessagingIntegration":
        integrations = cls.get_all(client=client)
        filtered_integrations = [
            integration for integration in integrations if integration.id == id
        ]
        if len(filtered_integrations) == 0:
            raise RuntimeError(f"Integration with ID {id} not found")
        return filtered_integrations[0]

    def update(self, **kwargs):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    @staticmethod
    def delete_by_id(client: Client, id: str):
        raise NotImplementedError()



import json
from typing import Any, Dict, Optional, Tuple, Union

from aporia.as_code.resources.base import BaseResource, CompareStatus, NonDeletableResourceException
from aporia.as_code.resources.data_source import DataSource
from aporia.sdk.client import Client
from aporia.sdk.data_sources import DataSource as _DataSource
from aporia.sdk.datasets import Dataset as _Dataset
from aporia.sdk.datasets import DatasetSchema, DatasetType, FieldSchema, FieldType, SortOrder


class Dataset(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        type: Union[DatasetType, str],
        connection_data: Dict[str, Any],
        id_column: Optional[str] = None,
        timestamp_column: Optional[str] = None,
        order_by_column: Optional[str] = None,
        group_by_column: Optional[str] = None,
        group_order_sort: Optional[Union[SortOrder, str]] = None,
        raw_inputs: Optional[Dict[str, Union[FieldType, str]]] = None,
        features: Optional[Dict[str, Union[FieldType, str]]] = None,
        predictions: Optional[Dict[str, Union[FieldType, str]]] = None,
        actuals: Optional[Dict[str, Union[FieldType, str]]] = None,
        actual_mappings: Optional[Dict[str, str]] = None,
        data_source_name: Optional[str] = None,
        data_source_id: Optional[str] = None,
        data_source: Optional[DataSource] = None,
    ):
        self.name = resource_name
        type = DatasetType(type)

        schema = DatasetSchema(
            id_column=id_column,
            timestamp_column=timestamp_column,
            group_by_column=group_by_column,
            order_by_column=order_by_column,
            group_order_sort=SortOrder(group_order_sort),
            raw_inputs=[
                FieldSchema(name=name, type=FieldType(type)) for name, type in raw_inputs.items()
            ]
            if raw_inputs is not None
            else [],
            features=[
                FieldSchema(name=name, type=FieldType(type)) for name, type in features.items()
            ]
            if features is not None
            else [],
            predictions=[
                FieldSchema(name=name, type=FieldType(type)) for name, type in predictions.items()
            ]
            if predictions is not None
            else [],
            actuals=[
                FieldSchema(
                    name=name,
                    type=FieldType(type),
                    properties={"prediction": actual_mappings[name]},
                )
                for name, type in actuals.items()
            ]
            if actuals is not None
            else [],
        )

        if data_source_id is None and data_source_name is None and data_source is None:
            raise ValueError("Must supply data source")

        if 1 != sum(
            [
                1
                for data_source_identifier in [
                    data_source,
                    data_source_id,
                    data_source_name,
                ]
                if data_source_identifier is not None
            ]
        ):
            raise ValueError("Must supply only one method to find data source")

        self._args = {
            "dataset_type": type,
            "connection_data": connection_data,
            "schema": schema,
        }

        if data_source is not None:
            data_source.dependants.append(
                (
                    self,
                    lambda data, dataset: dataset.setarg("data_source_id", data["id"]),
                )
            )
        elif data_source_id is not None:
            self._args["data_source_id"] = data_source_id
        else:

            def deferred_load(client: Client, resource: BaseResource):
                data_sources = _DataSource.get_all(client)
                same_named_data_sources = [
                    data_source
                    for data_source in data_sources
                    if data_source.name == data_source_name
                ]
                if len(same_named_data_sources) != 1:
                    raise Exception(
                        f"Found {len(same_named_data_sources)} data sources named {data_source_name}"
                    )
                data_source = same_named_data_sources[0]
                resource.setarg("data_source_id", data_source.id)

            self.deferred_load = deferred_load

    def _compare_schema(self, aporia_schema: Dict, compare_metadata: bool = True) -> bool:
        schema: DatasetSchema = self._args["schema"]

        raw_inputs = {
            raw_input.name: json.loads(raw_input.json()) for raw_input in schema.raw_inputs or []
        }
        features = {feature.name: json.loads(feature.json()) for feature in schema.features or []}
        predictions = {
            prediction.name: json.loads(prediction.json())
            for prediction in schema.predictions or []
        }
        actuals = {actual.name: json.loads(actual.json()) for actual in schema.actuals or []}

        schema_raw_inputs = {
            raw_input["name"]: raw_input for raw_input in aporia_schema["raw_inputs"] or []
        }
        schema_features = {feature["name"]: feature for feature in aporia_schema["features"] or []}
        schema_predictions = {
            prediction["name"]: prediction for prediction in aporia_schema["predictions"] or []
        }
        schema_actuals = {actual["name"]: actual for actual in aporia_schema["actuals"] or []}

        checks = [
            (schema.id_column == aporia_schema["id_column"]) if compare_metadata else True,
            (schema.timestamp_column == aporia_schema["timestamp_column"])
            if compare_metadata
            else True,
            raw_inputs == schema_raw_inputs,
            features == schema_features,
            predictions == schema_predictions,
            actuals == schema_actuals,
        ]

        return all(checks)

    def compare(self, resource_data: Dict) -> CompareStatus:
        if all(
            [
                self._args["dataset_type"].value == resource_data["stage"],
                self._args["connection_data"] == resource_data["config"],
                self._compare_schema(resource_data["schema"]),
                # TODO: Find a way to validate data source
                self._args["version_id"] == resource_data["model_version_id"],
            ]
        ):
            return CompareStatus.SAME
        elif any(
            [
                self._args["dataset_type"].value != resource_data["stage"],
                not self._compare_schema(resource_data["schema"], compare_metadata=False),
                self._args["version_id"] != resource_data["model_version_id"],
            ]
        ):
            return CompareStatus.MISMATCHED
        else:
            return CompareStatus.UPDATEABLE

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        dataset = _Dataset.create(client=client, **self._args)
        return dataset.id, dataset.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _Dataset.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        dataset = _Dataset.read(client=client, id=id)
        dataset.update(**self._args)
        return dataset.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        raise NonDeletableResourceException()

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        if self._args["connection_data"] != resource_data["config"]:
            diffs["connection_data"] = (resource_data["config"], self._args["connection_data"])
        return diffs



from typing import Any, cast, Dict, List, Optional, Tuple, Union

from aporia.as_code.resources.base import BaseResource, CompareStatus
from aporia.as_code.resources.custom_metrics import CustomMetric
from aporia.as_code.resources.dataset import Dataset
from aporia.as_code.resources.version import Version
from aporia.sdk.client import Client
from aporia.sdk.datasets import DatasetSchema, FieldType
from aporia.sdk.messaging import MessagingIntegrationType
from aporia.sdk.monitors import (
    AverageMethod,
    BaselineConfiguration,
    create_monitor_configuration,
    DetectionMethod,
    FocalConfiguration,
    MetricType,
)
from aporia.sdk.monitors import Monitor as _Monitor
from aporia.sdk.monitors import (
    MonitorConfiguration,
    MonitorType,
    SegmentIdentification,
    Severity,
    ThresholdConfiguration,
    TimePeriod,
)
from aporia.sdk.segments import Segment


def _is_feature_monitor(monitor_type: MonitorType) -> bool:
    return monitor_type in [
        MonitorType.DATA_DRIFT,
        MonitorType.MISSING_VALUES,
        MonitorType.METRIC_CHANGE,
        MonitorType.NEW_VALUES,
        MonitorType.VALUES_RANGE,
    ]


def _is_prediction_monitor(monitor_type: MonitorType) -> bool:
    return monitor_type in [
        MonitorType.PREDICTION_DRIFT,
        MonitorType.MISSING_VALUES,
        MonitorType.PERFORMANCE_DEGRADATION,
        MonitorType.VALUES_RANGE,
        MonitorType.NEW_VALUES,
    ]


class Monitor(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        # Monitoring Types
        monitor_type: Union[MonitorType, str],
        detection_method: Union[DetectionMethod, str],
        # Dataset Parameters
        focal: FocalConfiguration,
        # Alert parameters
        severity: Union[Severity, str],
        # Optional Monitor Parameters
        scheduling: Optional[str] = None,
        comment: Optional[str] = None,
        is_active: bool = True,
        creator: Optional[str] = None,
        version: Optional[Version] = None,
        version_id: Optional[str] = None,
        segment: Optional[Segment] = None,
        segment_id: Optional[str] = None,
        segment_value: Optional[Any] = None,
        # Optional Dataset Parameters
        baseline: Optional[BaselineConfiguration] = None,
        baseline_segment: Optional[Segment] = None,
        dataset: Optional[Dataset] = None,
        raw_inputs: Optional[List[str]] = None,
        features: Optional[List[str]] = None,
        predictions: Optional[List[str]] = None,
        is_embedding_monitor: bool = False,
        min_focal_prediction_count: Optional[int] = None,
        min_baseline_prediction_count: Optional[int] = None,
        # Optional Logic Evaluation Parameters
        percentage: Optional[int] = None,
        thresholds: Optional[ThresholdConfiguration] = None,
        sensitivity: Optional[float] = None,
        alert_on_increase_only: Optional[bool] = None,
        min: Optional[int] = None,
        max: Optional[int] = None,
        staleness_period: Optional[TimePeriod] = None,
        distance: Optional[float] = None,
        new_values_ratio_threshold: Optional[float] = None,
        new_values_count_threshold: Optional[int] = None,
        # Optional Metric Parameters
        metric: Optional[Union[MetricType, str]] = None,
        k: Optional[int] = None,
        prediction_class: Optional[Any] = None,
        prediction_threshold: Optional[float] = None,
        average_method: Optional[Union[AverageMethod, str]] = None,
        custom_metric: Optional[CustomMetric] = None,
        custom_metric_id: Optional[str] = None,
        # Optional Alert Parameters
        messaging: Optional[Dict[MessagingIntegrationType, List[str]]] = None,
        emails: Optional[List[str]] = None,
        max_alerts_per_day: Optional[int] = None,
        # Optional rename
        name: Optional[str] = None,
    ):
        self.name = resource_name
        if name is None:
            name = resource_name

        monitor_type = MonitorType(monitor_type)
        detection_method = DetectionMethod(detection_method)
        severity = Severity(severity)

        # TODO: Add by types
        use_predictions = _is_prediction_monitor(monitor_type)
        use_features = _is_feature_monitor(monitor_type)

        if dataset is not None:
            dataset_schema = cast(DatasetSchema, dataset._args["schema"])
            if len(dataset_schema.raw_inputs) > 0:
                raw_inputs = [
                    raw_input.name
                    for raw_input in dataset_schema.raw_inputs
                    if (raw_input.type is FieldType.EMBEDDING and is_embedding_monitor)
                    or (raw_input.type is not FieldType.EMBEDDING and not is_embedding_monitor)
                ]
            if len(dataset_schema.features) > 0:
                features = [
                    feature.name
                    for feature in dataset_schema.features
                    if (feature.type is FieldType.EMBEDDING and is_embedding_monitor)
                    or (feature.type is not FieldType.EMBEDDING and not is_embedding_monitor)
                ]
            if len(dataset_schema.predictions) > 0:
                predictions = [
                    prediction.name
                    for prediction in dataset_schema.predictions
                    if (prediction.type is FieldType.EMBEDDING and is_embedding_monitor)
                    or (prediction.type is not FieldType.EMBEDDING and not is_embedding_monitor)
                ]

        monitor_configuration = create_monitor_configuration(
            monitor_type=monitor_type,
            detection_method=detection_method,
            focal=focal,
            severity=severity,
            baseline=baseline,
            raw_inputs=raw_inputs if use_features else None,
            features=features if use_features else None,
            predictions=predictions if use_predictions else None,
            is_embedding_monitor=is_embedding_monitor,
            metric=metric,
            percentage=percentage,
            thresholds=thresholds,
            sensitivity=sensitivity,
            alert_on_increase_only=alert_on_increase_only,
            min=min,
            max=max,
            staleness_period=staleness_period,
            min_focal_prediction_count=min_focal_prediction_count,
            min_baseline_prediction_count=min_baseline_prediction_count,
            custom_metric_id=custom_metric.name
            if custom_metric is not None
            else custom_metric_id
            if custom_metric_id is not None
            else None,
            distance=distance,
            new_values_count_threshold=new_values_count_threshold,
            new_values_ratio_threshold=new_values_ratio_threshold,
            messaging=messaging,
            segment_identification=SegmentIdentification(group="PLACEHOLDER")
            if segment is not None
            else SegmentIdentification(group=segment_id, value=segment_value)
            if segment_id is not None
            else None,
            version=version_id if version_id is not None else None,
            emails=emails,
            max_alerts_per_day=max_alerts_per_day,
            k=k,
            prediction_class=prediction_class,
            prediction_threshold=prediction_threshold,
            average_method=AverageMethod(average_method) if average_method is not None else None,
        )

        # TODO: Test these things. Also with update
        if segment is not None:
            segment.dependants.append(
                (
                    self,
                    lambda data, monitor: monitor.setarg("segment_id", data["id"]),
                )
            )
        if baseline_segment is not None:
            baseline_segment.dependants.append(
                (
                    self,
                    lambda data, monitor: monitor.setarg("baseline_segment_id", data["id"]),
                )
            )
        if version is not None:
            version.dependants.append(
                (self, lambda data, monitor: monitor.setarg("version_id", data["id"]))
            )
        if custom_metric is not None:
            custom_metric.dependants.append(
                (self, lambda data, monitor: monitor.setarg("custom_metric_id", data["id"]))
            )

        self._args = {
            "name": name,
            "monitor_type": monitor_type,
            "configuration": monitor_configuration,
            "is_active": is_active,
        }

        if scheduling is not None:
            self._args["scheduling"] = scheduling
        if comment is not None:
            self._args["comment"] = comment
        if creator is not None:
            self._args["creator"] = creator

    def setarg(self, arg_name: str, arg_value: Any):
        # Model ID is added to the configuration by the SDK
        if arg_name == "segment_id":
            monitor_configuration = cast(MonitorConfiguration, self._args["configuration"])
            monitor_configuration.identification.segment.group = arg_value
            return
        if arg_name == "baseline_segment_id":
            monitor_configuration = cast(MonitorConfiguration, self._args["configuration"])
            monitor_configuration.baseline.segmentGroupId = arg_value
            return
        if arg_name == "version_id":
            monitor_configuration = cast(MonitorConfiguration, self._args["configuration"])
            monitor_configuration.identification.models.version = arg_value
            return
        if arg_name == "custom_metric_id":
            monitor_configuration = cast(MonitorConfiguration, self._args["configuration"])
            monitor_configuration.metric.id = arg_value
            return
        self._args[arg_name] = arg_value

    def compare(self, resource_data: Dict) -> CompareStatus:
        if self._args["configuration"].identification.models.id == "":
            self._args["configuration"].identification.models.id = self._args["model_id"]
        if all(
            [
                self._args["name"] == resource_data["name"],
                self._args.get("comment") == resource_data["comment"],
                self._args["is_active"] == resource_data["is_active"],
                self._args["configuration"].to_dict() == resource_data["configuration"],
            ]
        ):
            return CompareStatus.SAME
        elif any(
            [
                self._args["model_id"]
                != resource_data["configuration"]["identification"]["models"]["id"],
            ]
        ):
            return CompareStatus.MISMATCHED
        else:
            return CompareStatus.UPDATEABLE

    def create(self, client: Client) -> Tuple[str, Dict]:
        monitor = _Monitor.create(client=client, **self._args)
        return monitor.id, monitor.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _Monitor.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        monitor = _Monitor.read(client=client, id=id)
        monitor.update(**self._args)
        return monitor.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        _Monitor.delete_by_id(client=client, id=id)

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        if self._args["name"] != resource_data["name"]:
            diffs["name"] = (resource_data["name"], self._args["name"])
        if "comment" in self._args:
            if self._args["comment"] != resource_data["comment"]:
                diffs["comment"] = (resource_data["comment"], self._args["comment"])
        if self._args["is_active"] != resource_data["is_active"]:
            diffs["is_active"] = (resource_data["is_active"], self._args["is_active"])
        if self._args["configuration"].to_dict() != resource_data["configuration"]:
            diffs["configuration"] = (resource_data["configuration"], self._args["configuration"].to_dict())
        return diffs



from typing import Any, Dict, List, Optional, Tuple, Union

from aporia.as_code.resources.base import BaseResource, CompareStatus
from aporia.sdk.client import Client
from aporia.sdk.segments import Segment as _Segment


class Segment(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        name: Optional[str] = None,
        field: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        terms: Optional[List[Tuple[str, str]]] = None,
    ):
        self.name = resource_name
        self.dependants = []
        if name is None:
            name = resource_name

        self._args = {"name": name}
        if field is not None:
            if values is None:
                raise Exception("Must supply values for automatic segment")
            self._args["field_name"] = field
            self._args["values"] = values
        elif terms is not None:
            if values is not None:
                raise Exception("For custom segments, only specify the terms parameter")
            self._args["terms"] = terms
        else:
            raise Exception("Supply either field+values or terms")

    def compare(self, resource_data: Dict) -> CompareStatus:
        values = self._args.get("values", [])
        resource_field_name = resource_data.get("field_name")
        # Old stacks might have the field data stored in an outdated format (due to API change)
        if resource_field_name is None:
            resource_field_data = resource_data.get("field", {})
            resource_field_name = (resource_field_data or {}).get("name")

        if "terms" in self._args.keys():
            values = [term[1] for term in self._args["terms"]]
        if all(
            [
                self._args.get("terms") == resource_data["terms_values"],
                self._args.get("field_name") == resource_field_name,
                values == resource_data["values"],
                self._args["name"] == resource_data["name"],
            ]
        ):
            return CompareStatus.SAME
        elif any(
            [
                self._args.get("terms") != resource_data["terms_values"],
                self._args.get("field_name") != (resource_data["field"] or {}).get("name"),
                values != resource_data["values"],
            ]
        ):
            return CompareStatus.MISMATCHED
        else:
            return CompareStatus.UPDATEABLE

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        segment = _Segment.create(client=client, **self._args)
        return segment.id, segment.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _Segment.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        segment = _Segment.read(client=client, id=id)
        segment.update(**self._args)
        return segment.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        _Segment.delete_by_id(client=client, id=id)

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        if self._args["name"] != resource_data["name"]:
            diffs["name"] = (resource_data["name"], self._args["name"])
        return diffs